from collections import OrderedDict
from concurrent.futures import as_completed, ThreadPoolExecutor
from contextlib import contextmanager
import decimal
import inspect
import logging
import random
from sqlite3 import connect, Row
import threading
import time
import uuid

from orderedset import OrderedSet
from pymysql import escape_string
import pandas as pd
import sqlalchemy as sa
from stopit import TimeoutException, async_raise
from tlbx import (
    dbg,
    dbgsql,
    info,
    warn,
    error,
    get_class_var_values,
    sqlformat,
    json,
    st,
    chunks,
    initializer,
    is_int,
    orderedsetify,
    PrintMixin,
)

from zillion.configs import zillion_config
from zillion.core import (
    UnsupportedGrainException,
    UnsupportedKillException,
    ReportException,
    FailedKillException,
    ExecutionKilledException,
    ExecutionLockException,
    ExecutionState,
    AggregationTypes,
    DataSourceQueryModes,
    DataSourceQueryTimeoutException,
    FieldTypes,
    TechnicalTypes,
)
from zillion.field import get_table_fields, get_table_field_column, FormulaField
from zillion.sql_utils import sqla_compile, get_sqla_clause, to_sqlite_type

logging.getLogger(name="stopit").setLevel(logging.ERROR)

if zillion_config["DEBUG"]:
    logging.getLogger().setLevel(logging.DEBUG)

# Last unicode char - this helps get the rollup rows to sort last, but may
# need to be replaced for presentation
ROLLUP_INDEX_LABEL = chr(1114111)
ROLLUP_INDEX_PRETTY_LABEL = "::"
ROLLUP_TOTALS = "totals"

ROW_FILTER_OPS = [">", ">=", "<", "<=", "==", "!=", "in", "not in"]

PANDAS_ROLLUP_AGGR_TRANSLATION = {
    AggregationTypes.AVG: "mean",
    AggregationTypes.COUNT: "sum",
    AggregationTypes.COUNT_DISTINCT: "sum",
}

zillion_engine = sa.create_engine(zillion_config["ZILLION_DB_URL"])
zillion_metadata = sa.MetaData()
zillion_metadata.bind = zillion_engine

Reports = sa.Table(
    "reports",
    zillion_metadata,
    sa.Column("id", sa.Integer, primary_key=True),
    sa.Column("params", sa.Text),
    sa.Column("created_at", sa.DateTime, server_default=sa.func.now()),
)
zillion_metadata.create_all(zillion_engine)


class ExecutionStateMixin:
    def __init__(self):
        self._lock = threading.RLock()
        self._state = None

    @property
    def ready(self):
        return self._state == ExecutionState.READY

    @property
    def querying(self):
        return self._state == ExecutionState.QUERYING

    @property
    def killed(self):
        return self._state == ExecutionState.KILLED

    @contextmanager
    def get_lock(self, timeout=None):
        timeout = timeout or -1  # convert to `acquire` default if falsey

        result = self._lock.acquire(timeout=timeout)
        if not result:
            raise ExecutionLockException("lock wait timeout after %.3fs" % timeout)

        try:
            yield
        finally:
            self._lock.release()

    def raise_if_killed(self, timeout=None):
        with self.get_lock(timeout=timeout):
            if self.killed:
                raise ExecutionKilledException

    def get_state(self):
        return self._state

    def set_state(
        self,
        state,
        timeout=None,
        assert_ready=False,
        raise_if_killed=False,
        set_if_killed=False,
    ):
        assert state in get_class_var_values(ExecutionState), (
            "Invalid state value: %s" % state
        )
        cls_name = self.__class__.__name__

        with self.get_lock(timeout=timeout):
            if assert_ready:
                assert self.ready, "%s: expected ready state, got: %s" % (
                    cls_name,
                    self._state,
                )

            if raise_if_killed:
                try:
                    self.raise_if_killed()
                except ExecutionKilledException:
                    if set_if_killed:
                        dbg(
                            "%s: state transition: %s -> %s"
                            % (cls_name, self._state, state)
                        )
                        self._state = state
                    raise

            dbg("%s: state transition: %s -> %s" % (cls_name, self._state, state))
            self._state = state


class DataSourceQuery(ExecutionStateMixin, PrintMixin):
    repr_attrs = ["metrics", "dimensions", "criteria"]

    @initializer
    def __init__(self, warehouse, metrics, dimensions, criteria, table_set):
        self._conn = None
        self.field_map = {}
        self.metrics = metrics or {}
        self.dimensions = dimensions or {}
        self.select = self.build_select()
        super().__init__()
        self.set_state(ExecutionState.READY)

    def get_datasource_name(self):
        return self.table_set.datasource.name

    def get_datasource(self):
        return self.table_set.datasource

    def format_query(self):
        return sqlformat(sqla_compile(self.select))

    def get_bind(self):
        ds = self.get_datasource()
        assert ds.metadata.bind, (
            'Datasource "%s" does not have metadata.bind set' % ds.name
        )
        return ds.metadata.bind

    def get_dialect_name(self):
        return self.get_bind().dialect.name

    def get_conn(self):
        bind = self.get_bind()
        conn = bind.connect()
        return conn

    def execute(self, timeout=None, label=None):
        start = time.time()
        is_timeout = False
        t = None

        self.set_state(ExecutionState.QUERYING, assert_ready=True)

        try:
            assert not self._conn, "Called execute with active query connection"
            self._conn = self.get_conn()

            if label:
                self.select = self.select.comment(label)

            try:
                info("\n" + self.format_query())

                def do_timeout(main_thread):
                    nonlocal is_timeout
                    is_timeout = True
                    self.kill(main_thread=main_thread)

                if timeout:
                    main_thread = threading.current_thread()
                    t = threading.Timer(timeout, lambda: do_timeout(main_thread))
                    t.start()

                try:
                    result = self._conn.execute(self.select)
                    data = result.fetchall()
                except Exception as e:
                    if not is_timeout:
                        raise

                    diff = time.time() - start
                    self._conn.invalidate()
                    raise DataSourceQueryTimeoutException(
                        "query timed out after %.3fs" % diff
                    )
                finally:
                    if t:
                        t.cancel()
            finally:
                try:
                    self._conn.close()
                except Exception as e:
                    warn("Exception on connection close: %s" % str(e))
                self._conn = None

            diff = time.time() - start
            info("Got %d rows in %.3fs" % (len(data), diff))
            return DataSourceQueryResult(self, data, diff)
        finally:
            raise_if_killed = False if is_timeout else True
            self.set_state(
                ExecutionState.READY,
                raise_if_killed=raise_if_killed,
                set_if_killed=True,
            )

    def kill(self, main_thread=None):
        # Note: if query finishes as this is processing, it should end up
        # hitting raise_if_killed once this code is done rather than causing
        # the underlying connection to die and raise an exception.
        with self.get_lock():
            if self.ready:
                warn("kill called on query that isn't running")
                return

            if self.killed:
                warn("kill called on query already being killed")
                return

            self.set_state(ExecutionState.KILLED)

        # I don't see how this could happen, but get loud if it does...
        assert self._conn, "Attempting to kill with no active query connection"
        raw_conn = self._conn.connection

        dialect = self.get_dialect_name()
        info("Attempting kill on %s conn: %s" % (dialect, self._conn))

        # TODO: add support for more connection libraries
        if dialect == "mysql" and callable(getattr(raw_conn, "thread_id", None)):
            kill_conn = self.get_conn()
            conn_id = raw_conn.thread_id()
            try:
                kill_conn.execute("kill {}".format(conn_id))
            finally:
                kill_conn.close()
        elif dialect == "sqlite" and callable(getattr(raw_conn, "interrupt", None)):
            raw_conn.interrupt()
        elif dialect == "postgresql" and callable(getattr(raw_conn, "cancel", None)):
            raw_conn.cancel()  # TODO: assumes psycopg2
        elif main_thread:
            # This isn't guaranteed to work as the thread may be waiting for
            # an external resource to finish, but worth a shot.
            info("Trying async raise for unsupported dialect=%s" % dialect)
            async_raise(main_thread.ident, ExecutionKilledException)
        else:
            raise UnsupportedKillException("No kill support for dialect=%s" % dialect)

    def build_select(self):
        # https://docs.sqlalchemy.org/en/latest/core/selectable.html
        select = sa.select()

        join = self.get_join()
        select = select.select_from(join)

        for dimension in self.dimensions:
            select = select.column(self.get_field_expression(dimension))

        for metric in self.metrics:
            select = select.column(self.get_field_expression(metric))

        select = self.add_where(select)
        select = self.add_group_by(select)
        return select

    def get_field(self, name):
        if name in self.metrics:
            return self.metrics[name]
        elif name in self.dimensions:
            return self.dimensions[name]
        for row in self.criteria:
            if row[0].name == name:
                return row[0]
        assert False, "Could not find field for DataSourceQuery: %s" % name

    def column_for_field(self, field, table=None):
        ts = self.table_set

        if table is not None:
            column = get_table_field_column(
                ts.datasource.get_table(table.fullname), field
            )
        else:
            if ts.join and field in ts.join.field_map:
                column = ts.join.field_map[field]
            elif field in get_table_fields(
                ts.datasource.get_table(ts.ds_table.fullname)
            ):
                column = self.column_for_field(field, table=ts.ds_table)
            else:
                assert False, "Could not determine column for field %s" % field

        self.field_map[field] = column
        return column

    def get_field_expression(self, field, label=True):
        column = self.column_for_field(field)
        field_obj = self.get_field(field)
        return field_obj.get_ds_expression(column, label=label)

    def get_join(self):
        ts = self.table_set
        sqla_join = None
        last_table = None

        if not ts.join:
            return ts.ds_table

        for join_part in ts.join.join_parts:
            for table_name in join_part.table_names:
                table = ts.datasource.get_table(table_name)

                if sqla_join is None:
                    sqla_join = table
                    last_table = table
                    continue

                if table == last_table:
                    continue

                conditions = []
                for field in join_part.join_fields:
                    last_column = self.column_for_field(field, table=last_table)
                    column = self.column_for_field(field, table=table)
                    conditions.append(column == last_column)
                sqla_join = sqla_join.outerjoin(table, *tuple(conditions))
                last_table = table

        return sqla_join

    def add_where(self, select):
        if not self.criteria:
            return select
        for row in self.criteria:
            field = row[0]
            expr = self.get_field_expression(field.name, label=False)
            clause = sa.and_(get_sqla_clause(expr, row))
            select = select.where(clause)
        return select

    def add_group_by(self, select):
        if not self.dimensions:
            return select
        return select.group_by(
            *[sa.text(str(x)) for x in range(1, len(self.dimensions) + 1)]
        )

    def add_order_by(self, select, asc=True):
        if not self.dimensions:
            return select
        order_func = sa.asc
        if not asc:
            order_func = sa.desc
        return select.order_by(*[order_func(sa.text(x)) for x in self.dimensions])

    def covers_metric(self, metric):
        if metric in self.table_set.get_covered_metrics(self.warehouse):
            return True
        return False

    def covers_field(self, field):
        if field in self.table_set.get_covered_fields():
            return True
        return False

    def add_metric(self, name):
        assert self.covers_metric(name), "Metric %s can not be covered by query" % name
        # TODO: improve the way we maintain targeted metrics/dims
        self.table_set.target_fields.add(name)
        self.metrics[name] = self.table_set.datasource.get_metric(name)
        self.select = self.select.column(self.get_field_expression(name))


class DataSourceQuerySummary(PrintMixin):
    repr_attrs = ["datasource_name", "rowcount", "duration"]

    def __init__(self, query, data, duration):
        self.datasource_name = query.get_datasource_name()
        self.metrics = query.metrics
        self.dimensions = query.dimensions
        self.select = query.select
        self.duration = round(duration, 4)
        self.rowcount = len(data)

    def format_query(self):
        return sqlformat(sqla_compile(self.select))

    def format(self):
        sql = self.format_query()
        parts = [
            "%d rows in %.4f seconds" % (self.rowcount, self.duration),
            "Datasource: %s" % self.datasource_name,
            "Metrics: %s" % list(self.metrics),
            "Dimensions: %s" % list(self.dimensions),
            "\n%s" % sql,
            # TODO: Explain of query plan
        ]
        return "\n".join(parts)


class DataSourceQueryResult(PrintMixin):
    repr_attrs = ["summary"]

    def __init__(self, query, data, duration):
        self.query = query
        self.data = data
        self.summary = DataSourceQuerySummary(query, data, duration)


class BaseCombinedResult:
    @initializer
    def __init__(
        self, warehouse, ds_query_results, primary_ds_dimensions, adhoc_datasources=None
    ):
        self.conn = self.get_conn()
        self.cursor = self.get_cursor(self.conn)
        self.adhoc_datasources = adhoc_datasources or []
        self.table_name = "zillion_%s_%s" % (
            str(time.time()).replace(".", "_"),
            random.randint(0, 1e9),
        )
        self.primary_ds_dimensions = (
            orderedsetify(primary_ds_dimensions) if primary_ds_dimensions else []
        )
        self.ds_dimensions, self.ds_metrics = self.get_fields()
        self.create_table()
        self.load_table()

    def get_conn(self):
        raise NotImplementedError

    def get_cursor(self, conn):
        raise NotImplementedError

    def create_table(self):
        raise NotImplementedError

    def load_table(self):
        raise NotImplementedError

    def clean_up(self):
        raise NotImplementedError

    def get_final_result(self, metrics, dimensions, row_filters, rollup, pivot):
        raise NotImplementedError

    def get_row_hash(self, row):
        return hash(row[: len(self.primary_ds_dimensions)])

    def get_fields(self):
        dimensions = OrderedDict()
        metrics = OrderedDict()

        for qr in self.ds_query_results:
            for dim_name, dim in qr.query.dimensions.items():
                if dim_name in dimensions:
                    continue
                dimensions[dim_name] = dim

            for metric_name, metric in qr.query.metrics.items():
                if metric_name in metrics:
                    continue
                metrics[metric_name] = metric

        return dimensions, metrics

    def get_field_names(self):
        dims, metrics = self.get_fields()
        return list(dims.keys()) + list(metrics.keys())


class SQLiteMemoryCombinedResult(BaseCombinedResult):
    def get_conn(self):
        return connect(":memory:")

    def get_cursor(self, conn):
        conn.row_factory = Row
        return conn.cursor()

    def create_table(self):
        create_sql = "CREATE TEMP TABLE %s (" % self.table_name
        column_clauses = ["hash BIGINT NOT NULL PRIMARY KEY"]

        for field_name, field in self.ds_dimensions.items():
            type_str = str(to_sqlite_type(field.type))
            clause = "%s %s NOT NULL" % (field_name, type_str)
            column_clauses.append(clause)

        for field_name, field in self.ds_metrics.items():
            type_str = str(to_sqlite_type(field.type))
            clause = "%s %s DEFAULT NULL" % (field_name, type_str)
            column_clauses.append(clause)

        create_sql += ", ".join(column_clauses)
        create_sql += ") WITHOUT ROWID"
        create_sql = escape_string(create_sql)
        dbg(create_sql)  # Creates don't pretty print well with dbgsql?

        self.cursor.execute(create_sql)
        if self.primary_ds_dimensions:
            index_sql = "CREATE INDEX idx_dims ON %s (%s)" % (
                self.table_name,
                ", ".join(self.primary_ds_dimensions),
            )
            index_sql = escape_string(index_sql)
            dbgsql(index_sql)
            self.cursor.execute(index_sql)
        self.conn.commit()

    def get_bulk_insert_sql(self, rows):
        columns = [k for k in rows[0].keys()]
        placeholder = "(%s)" % (", ".join(["?"] * (1 + len(columns))))
        columns_clause = "hash, " + ", ".join(columns)

        sql = "INSERT INTO %s (%s) VALUES %s" % (
            self.table_name,
            columns_clause,
            placeholder,
        )

        update_clauses = []
        for k in columns:
            if k in self.primary_ds_dimensions:
                continue
            update_clauses.append("%s=excluded.%s" % (k, k))

        if update_clauses:
            update_clause = " ON CONFLICT(hash) DO UPDATE SET " + ", ".join(
                update_clauses
            )
            sql = sql + update_clause

        values = []
        for row in rows:
            row_values = [self.get_row_hash(row)]
            for value in row.values():
                # XXX: sqlite cant handle Decimal values. Alternative approach:
                # https://stackoverflow.com/questions/6319409/how-to-convert-python-decimal-to-sqlite-numeric
                if isinstance(value, decimal.Decimal):
                    value = float(value)
                row_values.append(value)
            values.append(row_values)

        return escape_string(sql), values

    def load_table(self):
        for qr in self.ds_query_results:
            for rows in chunks(qr.data, zillion_config["LOAD_TABLE_CHUNK_SIZE"]):
                insert_sql, values = self.get_bulk_insert_sql(rows)
                self.cursor.executemany(insert_sql, values)
            self.conn.commit()

    def select_all(self):
        qr = self.cursor.execute("SELECT * FROM %s" % self.table_name)
        return [OrderedDict(row) for row in qr.fetchall()]

    def get_final_select_sql(self, columns, dimension_aliases):
        columns_clause = ", ".join(columns)
        order_clause = "1"
        if dimension_aliases:
            order_clause = ", ".join(["%s ASC" % d for d in dimension_aliases])
        sql = "SELECT %s FROM %s GROUP BY hash ORDER BY %s" % (
            columns_clause,
            self.table_name,
            order_clause,
        )
        info("\n" + sqlformat(sql))
        return sql

    def apply_row_filters(self, df, row_filters, metrics, dimensions):
        filter_parts = []
        for row_filter in row_filters:
            field, op, value = row_filter
            assert (field in metrics) or (field in dimensions), (
                'Row filter field "%s" is not in result table' % field
            )
            assert op in ROW_FILTER_OPS, "Invalid row filter operation: %s" % op
            filter_parts.append("(%s %s %s)" % (field, op, value))
        return df.query(" and ".join(filter_parts))

    def get_multi_rollup_df(self, df, rollup, dimensions, aggrs, wavg, wavgs):
        # TODO: signature of this is a bit odd
        # TODO: test weighted averages
        # https://stackoverflow.com/questions/36489576/why-does-concatenation-of-dataframes-get-exponentially-slower
        level_aggrs = [df]
        dim_names = list(dimensions.keys())

        for level in range(rollup):
            if (level + 1) == len(dimensions):
                # Unnecessary to rollup at the most granular level
                break

            grouped = df.groupby(level=list(range(0, level + 1)))
            level_aggr = grouped.agg(aggrs, skipna=True)
            for metric_name, weighting_metric in wavgs:
                level_aggr[metric_name] = wavg(metric_name, weighting_metric)

            # for the remaining levels, set index cols to ROLLUP_INDEX_LABEL
            if level != (len(dimensions) - 1):
                new_index_dims = []
                for dim in dim_names[level + 1 :]:
                    level_aggr[dim] = ROLLUP_INDEX_LABEL
                    new_index_dims.append(dim)
                level_aggr = level_aggr.set_index(new_index_dims, append=True)

            level_aggrs.append(level_aggr)

        df = pd.concat(level_aggrs, sort=False, copy=False)
        df.sort_index(inplace=True)
        return df

    def apply_rollup(self, df, rollup, metrics, dimensions):
        assert dimensions, "Can not rollup without dimensions"

        aggrs = {}
        wavgs = []

        def wavg(avg_name, weight_name):
            d = df[avg_name]
            w = df[weight_name]
            try:
                return (d * w).sum() / w.sum()
            except ZeroDivisionError:
                return d.mean()  # Return mean if there are no weights

        for metric in metrics.values():
            if metric.weighting_metric:
                wavgs.append((metric.name, metric.weighting_metric))
                continue
            else:
                aggr_func = PANDAS_ROLLUP_AGGR_TRANSLATION.get(
                    metric.aggregation, metric.aggregation
                )
            aggrs[metric.name] = aggr_func

        aggr = df.agg(aggrs, skipna=True)
        for metric_name, weighting_metric in wavgs:
            aggr[metric_name] = wavg(metric_name, weighting_metric)

        apply_totals = True
        if rollup != ROLLUP_TOTALS:
            df = self.get_multi_rollup_df(df, rollup, dimensions, aggrs, wavg, wavgs)
            if rollup != len(dimensions):
                apply_totals = False

        if apply_totals:
            totals_rollup_index = (
                (ROLLUP_INDEX_LABEL,) * len(dimensions)
                if len(dimensions) > 1
                else ROLLUP_INDEX_LABEL
            )
            with pd.option_context("mode.chained_assignment", None):
                df.at[totals_rollup_index, :] = aggr

        return df

    def apply_technicals(self, df, technicals, rounding):
        for metric, tech in technicals.items():
            rolling = df[metric].rolling(
                tech.window, min_periods=tech.min_periods, center=tech.center
            )
            if tech.type == TechnicalTypes.MA:
                df[metric] = rolling.mean()
            elif tech.type == TechnicalTypes.SUM:
                df[metric] = rolling.sum()
            elif tech.type == TechnicalTypes.BOLL:
                ma = rolling.mean()
                std = rolling.std()
                upper = metric + "_upper"
                lower = metric + "_lower"
                if metric in rounding:
                    # This adds some extra columns for the bounds, so we use the same rounding
                    # as the root metric if applicable.
                    df[upper] = round(ma + 2 * std, rounding[metric])
                    df[lower] = round(ma - 2 * std, rounding[metric])
                else:
                    df[upper] = ma + 2 * std
                    df[lower] = ma - 2 * std
            else:
                assert False, "Invalid technical type: %s" % tech.type
        return df

    def get_final_result(self, metrics, dimensions, row_filters, rollup, pivot):
        start = time.time()
        columns = []
        dimension_aliases = []

        for dim in dimensions.values():
            columns.append(
                "%s as %s"
                % (
                    dim.get_final_select_clause(
                        self.warehouse, adhoc_fms=self.adhoc_datasources
                    ),
                    dim.name,
                )
            )
            dimension_aliases.append(dim.name)

        technicals = {}
        rounding = {}
        for metric in metrics.values():
            if metric.technical:
                technicals[metric.name] = metric.technical
            if metric.rounding is not None:
                rounding[metric.name] = metric.rounding
            columns.append(
                "%s as %s"
                % (
                    metric.get_final_select_clause(
                        self.warehouse, adhoc_fms=self.adhoc_datasources
                    ),
                    metric.name,
                )
            )

        sql = self.get_final_select_sql(columns, dimension_aliases)

        df = pd.read_sql(sql, self.conn, index_col=dimension_aliases or None)

        if row_filters:
            df = self.apply_row_filters(df, row_filters, metrics, dimensions)

        if technicals:
            df = self.apply_technicals(df, technicals, rounding)

        if rollup:
            df = self.apply_rollup(df, rollup, metrics, dimensions)

        if rounding:
            df = df.round(rounding)

        if pivot:
            df = df.unstack(pivot)

        dbg(df)
        dbg("Final result took %.3fs" % (time.time() - start))
        return df

    def clean_up(self):
        drop_sql = "DROP TABLE IF EXISTS %s " % self.table_name
        self.cursor.execute(drop_sql)
        self.conn.commit()
        self.conn.close()


class Report(ExecutionStateMixin):
    def __init__(
        self,
        warehouse,
        metrics=None,
        dimensions=None,
        criteria=None,
        row_filters=None,
        rollup=None,
        pivot=None,
        adhoc_datasources=None,
    ):
        start = time.time()
        self.id = None  # TODO: consider renaming spec_id or save_id
        self.uuid = uuid.uuid1()
        self.warehouse = warehouse

        if adhoc_datasources:
            self.warehouse.run_integrity_checks(adhoc_datasources=adhoc_datasources)

        self._requested_metrics = metrics
        self._requested_dimensions = dimensions
        self._requested_criteria = criteria

        self.metrics = self._get_fields_dict(
            metrics, FieldTypes.METRIC, adhoc_datasources=adhoc_datasources
        )
        self.dimensions = self._get_fields_dict(
            dimensions, FieldTypes.DIMENSION, adhoc_datasources=adhoc_datasources
        )

        assert (
            self.metrics or self.dimensions
        ), "One of metrics or dimensions must be specified for Report"

        self.criteria = self._populate_criteria_fields(
            criteria or [], adhoc_datasources=adhoc_datasources
        )
        self.row_filters = row_filters or []

        self.rollup = None
        if rollup is not None:
            assert dimensions, "Must specify dimensions in order to use rollup"
            if rollup != ROLLUP_TOTALS:
                assert is_int(rollup) and (0 < int(rollup) <= len(dimensions)), (
                    "Invalid rollup value: %s" % rollup
                )
                self.rollup = int(rollup)
            else:
                self.rollup = rollup

        self.pivot = pivot or []
        if pivot:
            assert set(self.pivot).issubset(
                set(self.dimensions)
            ), "Pivot columms must be a subset of dimensions"

        self.adhoc_datasources = adhoc_datasources or []

        self.ds_metrics = OrderedDict()
        self.ds_dimensions = OrderedDict()

        for metric in self.metrics.values():
            self.add_ds_fields(metric)

        for dim in self.dimensions.values():
            self.add_ds_fields(dim)

        self.queries = self.build_ds_queries()
        self.combined_query = None
        self.result = None

        super().__init__()
        self.set_state(ExecutionState.READY)
        dbg("Report init took %.3fs" % (time.time() - start))

    def get_params(self):
        used_datasources = list({q.get_datasource_name() for q in self.queries})
        datasources = [ds.get_params() for ds in self.warehouse.get_datasources()]
        return dict(
            kwargs=dict(
                metrics=self._requested_metrics,
                dimensions=self._requested_dimensions,
                criteria=self._requested_criteria,
                row_filters=self.row_filters,
                rollup=self.rollup,
                pivot=self.pivot,
            ),
            datasources=datasources,
            used_datasources=used_datasources,
        )

    def get_json(self):
        return json.dumps(self.get_params())

    def save(self):
        conn = zillion_engine.connect()
        try:
            result = conn.execute(Reports.insert(), params=self.get_json())
            report_id = result.inserted_primary_key[0]
            assert report_id, "No report ID found!"
        finally:
            conn.close()
        self.id = report_id
        return report_id

    @classmethod
    def load_params(cls, report_id):
        s = sa.select([Reports.c.params]).where(Reports.c.id == report_id)
        conn = zillion_engine.connect()
        try:
            result = conn.execute(s)
            row = result.fetchone()
            params = json.loads(row["params"])
            return params
        finally:
            conn.close()

    @classmethod
    def from_params(cls, warehouse, params, adhoc_datasources=None):
        used_dses = set(params.get("used_datasources", []))
        wh_dses = warehouse.get_datasource_names()
        all_dses = set(wh_dses) | set([x.name for x in (adhoc_datasources or [])])
        if not used_dses.issubset(all_dses):
            raise ReportException(
                "Report requires datasources that are not present: %s Found: %s"
                % (used_dses - all_dses, all_dses)
            )
        return Report(
            warehouse, **params["kwargs"], adhoc_datasources=adhoc_datasources
        )

    @classmethod
    def load(cls, warehouse, report_id, adhoc_datasources=None):
        params = cls.load_params(report_id)
        result = cls.from_params(warehouse, params, adhoc_datasources=adhoc_datasources)
        return result

    @classmethod
    def delete(cls, report_id):
        s = Reports.delete().where(Reports.c.id == report_id)
        conn = zillion_engine.connect()
        try:
            result = conn.execute(s)
        finally:
            conn.close()

    def _get_fields_dict(self, names, field_type, adhoc_datasources=None):
        d = OrderedDict()
        for name in names or []:
            if field_type == FieldTypes.METRIC:
                field = self.warehouse.get_metric(name, adhoc_fms=adhoc_datasources)
            elif field_type == FieldTypes.DIMENSION:
                field = self.warehouse.get_dimension(name, adhoc_fms=adhoc_datasources)
            else:
                assert False, "Invalid field type: %s" % field_type
            d[field.name] = field
        return d

    def _populate_criteria_fields(self, criteria, adhoc_datasources=None):
        field_names = [row[0] for row in criteria]
        fields_dict = self._get_fields_dict(
            field_names, FieldTypes.DIMENSION, adhoc_datasources=adhoc_datasources
        )
        final_criteria = []
        for row in criteria:
            row = list(row)
            field = fields_dict[row[0]]  # Replace field name with field object
            if isinstance(field, FormulaField):
                raise ReportException(
                    "FormulaFields are not allowed in criteria: %s" % field.name
                )
            row[0] = field
            final_criteria.append(row)
        return final_criteria

    def add_ds_fields(self, field):
        formula_fields, _ = field.get_formula_fields(
            self.warehouse, adhoc_fms=self.adhoc_datasources
        )
        if not formula_fields:
            formula_fields = [field.name]

        if field.field_type == FieldTypes.METRIC and field.weighting_metric:
            weighting_field = self.warehouse.get_metric(
                field.weighting_metric, adhoc_fms=self.adhoc_datasources
            )
            self.ds_metrics[field.weighting_metric] = weighting_field

        for formula_field in formula_fields:
            if formula_field == field.name:
                if field.field_type == FieldTypes.METRIC:
                    self.ds_metrics[formula_field] = field
                elif field.field_type == FieldTypes.DIMENSION:
                    self.ds_dimensions[formula_field] = field
                else:
                    assert False, "Invalid field_type: %s" % field.field_type
                continue

            if self.warehouse.has_metric(
                formula_field, adhoc_fms=self.adhoc_datasources
            ):
                self.ds_metrics[formula_field] = self.warehouse.get_metric(
                    formula_field, adhoc_fms=self.adhoc_datasources
                )
            elif self.warehouse.has_dimension(
                formula_field, adhoc_fms=self.adhoc_datasources
            ):
                self.ds_dimensions[formula_field] = self.warehouse.get_dimension(
                    formula_field, adhoc_fms=self.adhoc_datasources
                )
            else:
                assert False, "Could not find field %s in warehouse" % formula_field

    def get_query_label(self, query_label):
        return "Report: %s | Query: %s" % (str(self.uuid), query_label)

    def execute_ds_queries_sequential(self, queries):
        results = []
        timeout = zillion_config["DATASOURCE_QUERY_TIMEOUT"]
        for i, query in enumerate(queries):
            self.raise_if_killed()
            label = self.get_query_label("%s / %s" % (i + 1, len(queries)))
            result = query.execute(timeout=timeout, label=label)
            results.append(result)
        return results

    def execute_ds_queries_multithread(self, queries):
        # TODO: currently if any query times out, the entire report fails.
        # It might be better if partial results could be returned.
        finished = {}
        timeout = zillion_config["DATASOURCE_QUERY_TIMEOUT"]
        workers = zillion_config.get("DATASOURCE_QUERY_WORKERS", len(queries))

        with ThreadPoolExecutor(max_workers=workers) as executor:
            futures_map = {}
            for i, query in enumerate(queries):
                label = self.get_query_label("%s / %s" % (i + 1, len(queries)))
                future = executor.submit(query.execute, timeout=timeout, label=label)
                futures_map[future] = query

            for future in as_completed(futures_map):
                data = future.result()
                query = futures_map[future]
                finished[future] = data

        return finished.values()

    def execute_ds_queries(self, queries):
        mode = zillion_config["DATASOURCE_QUERY_MODE"]
        dbg("Executing %s datasource queries in %s mode" % (len(queries), mode))
        start = time.time()
        if mode == DataSourceQueryModes.SEQUENTIAL:
            return self.execute_ds_queries_sequential(queries)
        if mode == DataSourceQueryModes.MULTITHREAD:
            return self.execute_ds_queries_multithread(queries)
        dbg("DataSource queries took %.3fs" % (time.time() - start))
        assert False, "Invalid DATASOURCE_QUERY_MODE: %s" % mode

    def execute(self):
        start = time.time()

        self.set_state(ExecutionState.QUERYING, assert_ready=True)

        try:
            query_results = self.execute_ds_queries(self.queries)
            summaries = [x.summary for x in query_results]

            self.raise_if_killed()
            cr = self.create_combined_result(query_results)

            try:
                self.raise_if_killed()
                final_result = cr.get_final_result(
                    self.metrics,
                    self.dimensions,
                    self.row_filters,
                    self.rollup,
                    self.pivot,
                )
                diff = time.time() - start
                self.result = ReportResult(final_result, diff, summaries)
                return self.result
            finally:
                cr.clean_up()
        finally:
            self.set_state(
                ExecutionState.READY, raise_if_killed=True, set_if_killed=True
            )

    def kill(self, soft=False, raise_if_failed=False):
        info("killing report %s" % self.uuid)

        with self.get_lock():
            if self.ready:
                warn("kill called on report that isn't running")
                return

            if self.killed:
                warn("kill called on report already being killed")
                return

            if soft:
                self.set_state(ExecutionState.KILLED)
                return

            querying = self.querying  # grab state before changing it
            self.set_state(ExecutionState.KILLED)

            if querying:
                dbg("attempting kill on %d report queries" % len(self.queries))
                exceptions = []

                for query in self.queries:
                    try:
                        query.kill()
                    except Exception as e:
                        setattr(e, "query", query)
                        exceptions.append(e)

                if exceptions:
                    if raise_if_failed:
                        raise FailedKillException(exceptions)
                    else:
                        warn("failed to kill some queries: %s" % exceptions)

    def get_grain(self):
        if not (self.ds_dimensions or self.criteria):
            return None
        grain = set()
        if self.ds_dimensions:
            grain = grain | self.ds_dimensions.keys()
        if self.criteria:
            grain = grain | {x[0].name for x in self.criteria}
        return grain

    def build_ds_queries(self):
        grain = self.get_grain()
        grain_errors = []
        queries = []

        def metric_covered_in_queries(metric):
            for query in queries:
                if query.covers_metric(metric):
                    query.add_metric(metric)
                    return query
            return False

        for metric_name, metric in self.ds_metrics.items():
            existing_query = metric_covered_in_queries(metric_name)
            if existing_query:
                # TODO: we could do a single consolidation at the end instead
                # and that might get more optimal results
                dbg("Metric %s is covered by existing query" % metric_name)
                continue

            try:
                table_set = self.warehouse.get_metric_table_set(
                    metric_name, grain, adhoc_datasources=self.adhoc_datasources
                )
            except UnsupportedGrainException as e:
                # Gather all grain errors to be raised in one exception
                grain_errors.append(str(e))
                continue

            query = DataSourceQuery(
                self.warehouse,
                {metric.name: metric},
                self.ds_dimensions,
                self.criteria,
                table_set,
            )
            queries.append(query)

        if grain_errors:
            raise UnsupportedGrainException(grain_errors)

        if not self.ds_metrics:
            dbg("No metrics requested, getting dimension table sets")
            table_set = self.warehouse.get_dimension_table_set(
                grain, adhoc_datasources=self.adhoc_datasources
            )
            query = DataSourceQuery(
                self.warehouse, None, self.ds_dimensions, self.criteria, table_set
            )
            queries.append(query)

        for query in queries:
            dbgsql(sqla_compile(query.select))

        return queries

    def create_combined_result(self, ds_query_results):
        start = time.time()
        result = SQLiteMemoryCombinedResult(
            self.warehouse,
            ds_query_results,
            list(self.ds_dimensions.keys()),
            adhoc_datasources=self.adhoc_datasources,
        )
        dbg("Combined result took %.3fs" % (time.time() - start))
        return result


class ReportResult(PrintMixin):
    repr_attrs = ["rowcount", "duration", "query_summaries"]

    @initializer
    def __init__(self, df, duration, query_summaries):
        self.duration = round(duration, 4)
        self.rowcount = len(df)

    def get_rollup_mask(self):
        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Index.isin.html
        mask = None
        for i, level in enumerate(self.df.index.levels):
            if mask is None:
                mask = self.df.index.isin([ROLLUP_INDEX_LABEL], i)
            else:
                mask = mask | self.df.index.isin([ROLLUP_INDEX_LABEL], i)
        return mask

    def rollup_rows(self):
        return self.df.loc[self.get_rollup_mask()]

    def non_rollup_rows(self):
        return self.df.loc[~self.get_rollup_mask()]



from tlbx import ClassValueContainsMeta

ADHOC_URL = "adhoc"


class FieldTypes(metaclass=ClassValueContainsMeta):
    DIMENSION = "dimension"
    METRIC = "metric"


class TableTypes(metaclass=ClassValueContainsMeta):
    DIMENSION = "dimension"
    METRIC = "metric"


class AggregationTypes(metaclass=ClassValueContainsMeta):
    AVG = "avg"
    COUNT = "count"
    COUNT_DISTINCT = "count_distinct"
    MIN = "min"
    MAX = "max"
    SUM = "sum"


class TechnicalTypes(metaclass=ClassValueContainsMeta):
    MA = "MA"
    SUM = "SUM"
    BOLL = "BOLL"


class DataSourceQueryModes(metaclass=ClassValueContainsMeta):
    SEQUENTIAL = "sequential"
    MULTITHREAD = "multithread"


class ExecutionState:
    READY = "ready"
    QUERYING = "querying"
    KILLED = "killed"


class WarehouseException(Exception):
    pass


class ReportException(Exception):
    pass


class UnsupportedGrainException(Exception):
    pass


class UnsupportedKillException(Exception):
    pass


class FailedKillException(Exception):
    pass


class DataSourceQueryTimeoutException(Exception):
    pass


class ExecutionKilledException(Exception):
    pass


class ExecutionLockException(Exception):
    pass


class InvalidFieldException(Exception):
    pass


class DisallowedSQLException(Exception):
    pass


class MaxFormulaDepthException(Exception):
    pass



from collections import defaultdict, OrderedDict
import copy
import datetime
import logging
import random
import time

import networkx as nx
import pandas as pd
import sqlalchemy as sa
from tlbx import (
    dbg,
    info,
    warn,
    pf,
    st,
    rmfile,
    initializer,
    get_string_format_args,
    iter_or,
    powerset,
    PrintMixin,
    MappingMixin,
)

from zillion.configs import (
    AdHocFieldSchema,
    AdHocMetricSchema,
    WarehouseConfigSchema,
    ColumnInfo,
    TableInfo,
    MetricConfigSchema,
    TechnicalInfoSchema,
    DimensionConfigSchema,
    parse_technical_string,
    is_valid_field_name,
    zillion_config,
    DATASOURCE_ALLOWABLE_CHARS,
)
from zillion.core import (
    UnsupportedGrainException,
    InvalidFieldException,
    MaxFormulaDepthException,
    WarehouseException,
    TableTypes,
)
from zillion.datasource import DataSource, AdHocDataSource, datasource_from_config
from zillion.field import (
    Field,
    Metric,
    Dimension,
    Technical,
    FormulaMetric,
    create_metric,
    create_dimension,
    get_table_metrics,
    get_table_dimensions,
    get_table_fields,
    FieldManagerMixin,
)
from zillion.report import Report
from zillion.sql_utils import (
    infer_aggregation_and_rounding,
    aggregation_to_sqla_func,
    is_probably_metric,
    sqla_compile,
    column_fullname,
)

if zillion_config["DEBUG"]:
    logging.getLogger().setLevel(logging.DEBUG)


class Warehouse(FieldManagerMixin):
    def __init__(self, config=None, datasources=None, ds_priority=None):
        self.datasources = OrderedDict()
        self._metrics = {}
        self._dimensions = {}
        self._supported_dimension_cache = {}
        self._created_adhoc_datasources = set()

        for ds in datasources or []:
            self.add_datasource(ds, skip_integrity_checks=True)

        if config:
            config = WarehouseConfigSchema().load(config)
            self.apply_config(config, skip_integrity_checks=True)

        assert self.datasources, "No datasources provided or found in config"
        self.run_integrity_checks()

        self.ds_priority = ds_priority or list(self.datasources.keys())

        assert isinstance(self.ds_priority, list), (
            "Invalid format for ds_priority, must be list of datasource names: %s"
            % self.ds_priority
        )
        assert len(self.ds_priority) == len(
            self.datasources
        ), "Length mismatch between ds_priority and datasources"
        for ds_name in self.ds_priority:
            assert ds_name in self.datasources, (
                "Datasource %s is in ds_priority but not in datasource map" % ds_name
            )

    def __repr__(self):
        return "<%s> Datasources: %s" % (
            self.__class__.__name__,
            self.get_datasource_names(),
        )

    def clean_up(self):
        for ds_name, ds in self.datasources.items():
            if ds_name in self._created_adhoc_datasources:
                info("Cleaning up warehouse adhoc ds %s" % ds)
                ds.clean_up()

    def print_info(self):
        print("---- Warehouse")
        print("metrics:")
        self.print_metrics(indent=2)
        print("dimensions:")
        self.print_dimensions(indent=2)

        for ds in self.get_datasources():
            print()
            ds.print_info()

    def get_datasources(self):
        return self.datasources.values()

    def get_datasource_names(self):
        return self.datasources.keys()

    def get_datasource(self, name, adhoc_datasources=None):
        if name in self.datasources:
            return self.datasources[name]

        for adhoc_datasource in adhoc_datasources or []:
            if adhoc_datasource.name == name:
                return adhoc_datasource

        assert False, 'Could not find datasource with name "%s"' % name

    def get_child_field_managers(self):
        return list(self.get_datasources())

    def add_datasource(self, ds, skip_integrity_checks=False):
        dbg("Adding datasource %s" % ds.name)
        self.datasources[ds.name] = ds
        if not skip_integrity_checks:
            self.run_integrity_checks()

    def remove_datasource(self, ds):
        dbg("Removing datasource %s" % ds.name)
        del self.datasources[ds.name]

    def create_or_update_datasources(self, ds_configs, skip_integrity_checks=False):
        for ds_name in ds_configs:
            if ds_name in self.datasources:
                self.datasources[ds_name].apply_config(ds_configs[ds_name])
                continue

            ds = datasource_from_config(ds_name, ds_configs[ds_name])
            if isinstance(ds, AdHocDataSource):
                # We track this so we can clean up later
                self._created_adhoc_datasources.add(ds.name)
            self.add_datasource(ds, skip_integrity_checks=skip_integrity_checks)

    def apply_config(self, config, skip_integrity_checks=False):
        self.populate_global_fields(config, force=True)
        self.create_or_update_datasources(
            config.get("datasources", {}), skip_integrity_checks=skip_integrity_checks
        )

    def _check_conflicting_fields(self, adhoc_datasources=None):
        # TODO: in addition to checking metric vs dimension settings
        # we could add type comparisons (make sure its always num or str)
        errors = []

        for field in self.get_field_names(adhoc_fms=adhoc_datasources):
            if self.has_metric(
                field, adhoc_fms=adhoc_datasources
            ) and self.has_dimension(field, adhoc_fms=adhoc_datasources):
                errors.append("Field %s is in both metrics and dimensions" % field)

        return errors

    def _check_fields_have_type(self, adhoc_datasources=None):
        errors = []

        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            for table in ds.metadata.tables.values():
                if not table.zillion:
                    continue

                for column in table.c:
                    if not column.zillion:
                        continue

                    for field in column.zillion.get_field_names():
                        if not (
                            self.has_metric(field, adhoc_fms=adhoc_datasources)
                            or self.has_dimension(field, adhoc_fms=adhoc_datasources)
                        ):
                            errors.append(
                                "Field %s for column %s->%s is not defined as a metric or dimension"
                                % (field, ds.name, column_fullname(column))
                            )

        return errors

    def _check_primary_key_dimensions(self, adhoc_datasources=None):
        errors = []

        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            for table in ds.metadata.tables.values():
                if not table.zillion:
                    continue

                for column in table.c:
                    if not column.zillion:
                        continue

                    if column.primary_key:
                        # TODO: does this make sense for composite keys?
                        # Why is this even necessary?
                        dims = set()
                        for field in column.zillion.get_field_names():
                            if self.has_dimension(
                                field, adhoc_fms=adhoc_datasources
                            ) or not self.has_metric(
                                field, adhoc_fms=adhoc_datasources
                            ):
                                dims.add(field)
                            if len(dims) >= 2:
                                errors.append(
                                    "Primary key column %s may only map to a single dimension: %s"
                                    % (column, dims)
                                )

        return errors

    def _check_weighting_metrics(self, adhoc_datasources=None):
        errors = []

        for metric in self.get_metrics(adhoc_fms=adhoc_datasources).values():
            if not metric.weighting_metric:
                continue

            for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
                tables = ds.get_tables_with_field(metric.name)
                if not tables:
                    continue

                for table in tables:
                    if not metric.weighting_metric in get_table_fields(table):
                        errors.append(
                            "Table %s->%s has metric %s but not weighting metric %s"
                            % (
                                ds.name,
                                table.fullname,
                                metric.name,
                                metric.weighting_metric,
                            )
                        )

        return errors

    def run_integrity_checks(self, adhoc_datasources=None):
        errors = []
        if adhoc_datasources:
            for ds in adhoc_datasources:
                if ds.name in self.datasources:
                    errors.append(
                        "Adhoc DataSource '%s' name conflicts with existing DataSource"
                        % ds.name
                    )
        errors.extend(
            self._check_conflicting_fields(adhoc_datasources=adhoc_datasources)
        )
        errors.extend(self._check_fields_have_type(adhoc_datasources=adhoc_datasources))
        errors.extend(
            self._check_primary_key_dimensions(adhoc_datasources=adhoc_datasources)
        )
        errors.extend(
            self._check_weighting_metrics(adhoc_datasources=adhoc_datasources)
        )
        if errors:
            raise WarehouseException("Integrity check(s) failed.\n%s" % pf(errors))

    def get_supported_dimensions_for_metric(
        self, metric, use_cache=True, adhoc_datasources=None
    ):
        dims = set()
        metric = self.get_metric(metric, adhoc_fms=adhoc_datasources)

        if use_cache and metric.name in self._supported_dimension_cache:
            # XXX When to clear cache?
            return self._supported_dimension_cache[metric]

        for ds in self.get_datasources():
            ds_tables = ds.get_tables_with_field(metric.name)
            used_tables = set()

            for ds_table in ds_tables:
                if ds_table.fullname not in used_tables:
                    dims |= get_table_dimensions(
                        self, ds_table, adhoc_fms=adhoc_datasources
                    )
                    used_tables.add(ds_table.fullname)

                desc_tables = nx.descendants(ds.graph, ds_table.fullname)
                for desc_table in desc_tables:
                    if desc_table not in used_tables:
                        dims |= get_table_dimensions(
                            self, ds.get_table(desc_table), adhoc_fms=adhoc_datasources
                        )
                        used_tables.add(desc_table)

        self._supported_dimension_cache[metric] = dims
        return dims

    def get_supported_dimensions(self, metrics, adhoc_datasources=None):
        dims = set()
        for metric in metrics:
            supported_dims = self.get_supported_dimensions_for_metric(
                metric, adhoc_datasources=adhoc_datasources
            )
            dims = (dims & supported_dims) if len(dims) else supported_dims
        return dims

    def get_ds_tables_with_metric(self, metric, adhoc_datasources=None):
        ds_tables = defaultdict(list)
        count = 0
        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            tables = ds.get_tables_with_field(metric)
            if not tables:
                continue
            ds_tables[ds.name] = tables
            count += 1
        dbg(
            "found %d datasources, %d columns for metric %s"
            % (len(ds_tables), count, metric)
        )
        return ds_tables

    def get_ds_dim_tables_with_dim(self, dim, adhoc_datasources=None):
        ds_tables = defaultdict(list)
        count = 0
        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            tables = ds.get_dim_tables_with_dim(dim)
            if not tables:
                continue
            ds_tables[ds.name] = tables
            count += 1
        dbg(
            "found %d datasources, %d columns for dim %s" % (len(ds_tables), count, dim)
        )
        return ds_tables

    def get_ds_table_sets(self, ds_tables, field, grain, adhoc_datasources=None):
        """Returns all table sets that can satisfy grain in each datasource"""
        ds_table_sets = {}
        for ds_name, ds_tables_with_field in ds_tables.items():
            ds = self.get_datasource(ds_name, adhoc_datasources=adhoc_datasources)
            possible_table_sets = ds.find_possible_table_sets(
                ds_tables_with_field, field, grain
            )
            if not possible_table_sets:
                continue
            ds_table_sets[ds_name] = possible_table_sets
        dbg(ds_table_sets)
        return ds_table_sets

    def choose_best_data_source(self, ds_names):
        if self.ds_priority:
            for ds_name in self.ds_priority:
                if ds_name in ds_names:
                    return ds_name

        # TODO: eventually it would be nice to choose a datasource if:
        #  A) Its historically been faster
        #  B) All of the requested data can be pulled from one datasource
        warn("No datasource priorities established, choosing first option")
        assert ds_names, "No datasource names provided"
        return ds_names[0]

    def choose_best_table_set(self, ds_table_sets):
        ds_name = self.choose_best_data_source(list(ds_table_sets.keys()))
        if len(ds_table_sets[ds_name]) > 1:
            # TODO: establish table set priorities based on expected query performance?
            warn(
                "Picking smallest of %d available table sets"
                % len(ds_table_sets[ds_name])
            )
        return sorted(ds_table_sets[ds_name], key=lambda x: len(x))[0]

    def generate_unsupported_grain_msg(self, grain, metric, adhoc_datasources=None):
        """
        This assumes you are in a situation where you are sure the metric can not
        meet the grain and want to generate a helpful message pinpointing the
        issue.  If the metric actually supports all dimensions, the conclusion
        is that it just doesn't support them all in a single datasource and
        thus can't meet the grain.
        """
        supported = self.get_supported_dimensions_for_metric(
            metric, adhoc_datasources=adhoc_datasources
        )
        unsupported = grain - supported
        if unsupported:
            msg = (
                "metric %s can not meet grain %s due to unsupported dimensions: %s"
                % (metric, grain, unsupported)
            )
        else:
            msg = "metric %s can not meet grain %s in any single datasource" % (
                metric,
                grain,
            )
        return msg

    def get_metric_table_set(self, metric, grain, adhoc_datasources=None):
        dbg("metric:%s grain:%s" % (metric, grain))
        ds_metric_tables = self.get_ds_tables_with_metric(
            metric, adhoc_datasources=adhoc_datasources
        )
        ds_table_sets = self.get_ds_table_sets(
            ds_metric_tables, metric, grain, adhoc_datasources=adhoc_datasources
        )
        if not ds_table_sets:
            msg = self.generate_unsupported_grain_msg(
                grain, metric, adhoc_datasources=adhoc_datasources
            )
            raise UnsupportedGrainException(msg)
        table_set = self.choose_best_table_set(ds_table_sets)
        return table_set

    def get_dimension_table_set(self, grain, adhoc_datasources=None):
        """
        This is meant to be used in cases where no metrics are requested. We only
        allow it to look at dim tables since the assumption is joining to a metric
        table to explore dimensions doesn't make sense and would have poor performance.
        """
        dbg("grain:%s" % grain)

        table_set = None
        for dim_name in grain:
            ds_dim_tables = self.get_ds_dim_tables_with_dim(
                dim_name, adhoc_datasources=adhoc_datasources
            )
            ds_table_sets = self.get_ds_table_sets(
                ds_dim_tables, dim_name, grain, adhoc_datasources=adhoc_datasources
            )
            if not ds_table_sets:
                continue
            table_set = self.choose_best_table_set(ds_table_sets)
            break

        if not table_set:
            raise UnsupportedGrainException(
                "No dimension table set found to meet grain: %s" % grain
            )
        return table_set

    def load_report(self, report_id, adhoc_datasources=None):
        report = Report.load(self, report_id, adhoc_datasources=adhoc_datasources)
        return report

    def delete_report(self, report_id):
        report = Report.delete(report_id)
        return report

    def save_report(self, **kwargs):
        report = Report(self, **kwargs)
        report.save()
        return report

    def execute(
        self,
        metrics=None,
        dimensions=None,
        criteria=None,
        row_filters=None,
        rollup=None,
        pivot=None,
        adhoc_datasources=None,
    ):
        start = time.time()

        report = Report(
            self,
            metrics=metrics,
            dimensions=dimensions,
            criteria=criteria,
            row_filters=row_filters,
            rollup=rollup,
            pivot=pivot,
            adhoc_datasources=adhoc_datasources,
        )
        result = report.execute()

        dbg("warehouse report took %.3fs" % (time.time() - start))
        return result

    def execute_id(self, report_id, adhoc_datasources=None):
        start = time.time()
        report = self.load_report(report_id, adhoc_datasources=adhoc_datasources)
        result = report.execute()
        dbg("warehouse report took %.3fs" % (time.time() - start))
        return result



from collections import defaultdict, OrderedDict
import copy
import datetime
import random
from urllib.parse import urlparse, urlunparse, parse_qs

import pandas as pd
import networkx as nx
from orderedset import OrderedSet
import sqlalchemy as sa
from tlbx import (
    PrintMixin,
    dbg,
    pf,
    st,
    format_msg,
    rmfile,
    initializer,
    open_filepath_or_buffer,
    get_string_format_args,
)

from zillion.configs import (
    DATASOURCE_ALLOWABLE_CHARS,
    DATASOURCE_ALLOWABLE_CHARS_STR,
    TableInfo,
    ColumnInfo,
    DataSourceConfigSchema,
    TableConfigSchema,
    MetricConfigSchema,
    DimensionConfigSchema,
    default_field_name,
    zillion_config,
    ADHOC_TABLE_CONFIG_PARAMS,
    EXCLUDE,
)
from zillion.core import TableTypes, ADHOC_URL
from zillion.field import (
    Field,
    Metric,
    Dimension,
    Technical,
    FormulaMetric,
    create_metric,
    create_dimension,
    get_table_metrics,
    get_table_dimensions,
    get_table_fields,
    get_table_field_column,
    FieldManagerMixin,
)
from zillion.sql_utils import (
    column_fullname,
    infer_aggregation_and_rounding,
    get_dialect_type_conversions,
    is_probably_metric,
    get_schemas,
    filter_dialect_schemas,
)


class TableSet(PrintMixin):
    repr_attrs = ["datasource", "join", "grain", "target_fields"]

    @initializer
    def __init__(self, datasource, ds_table, join, grain, target_fields):
        self.adhoc_datasources = []
        if isinstance(self.datasource, AdHocDataSource):
            self.adhoc_datasources = [self.datasource]

    def get_covered_metrics(self, wh):
        # Note: we pass in WH instead of DS here since the metric/dim may only
        # be defined at the WH level even if it exists in the DS
        covered_metrics = get_table_metrics(
            wh, self.ds_table, adhoc_fms=self.adhoc_datasources
        )
        return covered_metrics

    def get_covered_fields(self):
        covered_fields = get_table_fields(
            self.ds_table, adhoc_fms=self.adhoc_datasources
        )
        return covered_fields

    def __len__(self):
        if not self.join:
            return 1
        return len(self.join.table_names)


class JoinPart(PrintMixin):
    repr_attrs = ["datasource", "table_names", "join_fields"]

    @initializer
    def __init__(self, datasource, table_names, join_fields):
        pass


class Join(PrintMixin):
    """
    Join is a group of join parts that would be used together.
    field_map represents the requested fields this join list is meant to satisfy
    """

    repr_attrs = ["datasource", "table_names", "field_map"]

    @initializer
    def __init__(self, join_parts, field_map):
        self.datasource = None
        self.table_names = OrderedSet()
        for join_part in self.join_parts:
            if not self.datasource:
                self.datasource = join_part.datasource
            else:
                assert join_part.datasource.name == self.datasource.name, (
                    "Can not form %s using join_parts from different datasources"
                    % self.__class__
                )
            for table_name in join_part.table_names:
                self.table_names.add(table_name)

    def __key(self):
        return tuple(self.table_names)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(self, type(other)) and self.__key() == other.__key()

    def __len__(self):
        return len(self.table_names)

    def get_covered_fields(self):
        """Generate a list of all possible fields this can cover"""
        fields = set()
        for table_name in self.table_names:
            table = self.datasource.metadata.tables[table_name]
            covered_fields = get_table_fields(table)
            fields = fields | covered_fields
        return fields

    def add_field(self, field):
        assert field not in self.field_map, "Field %s is already in field map" % field
        for table_name in self.table_names:
            table = self.datasource.metadata.tables[table_name]
            covered_fields = get_table_fields(table)
            if field in covered_fields:
                column = get_table_field_column(table, field)
                self.field_map[field] = column
                return
        assert False, "Field %s is not in any join tables: %s" % (
            field,
            self.table_names,
        )

    def add_fields(self, fields):
        for field in fields:
            if field not in self.field_map:
                self.add_field(field)


def joins_from_path(ds, path, field_map=None):
    join_parts = []
    if len(path) == 1:
        # A placeholder join that is really just a single table
        join_part = JoinPart(ds, path, None)
        join_parts.append(join_part)
    else:
        for i, node in enumerate(path):
            if i == (len(path) - 1):
                break
            start, end = path[i], path[i + 1]
            edge = ds.graph.edges[start, end]
            join_part = JoinPart(ds, [start, end], edge["join_fields"])
            join_parts.append(join_part)
    return Join(join_parts, field_map=field_map)


class NeighborTable(PrintMixin):
    repr_attrs = ["table_name", "join_fields"]

    @initializer
    def __init__(self, table, join_fields):
        self.table_name = table.fullname


class DataSource(FieldManagerMixin, PrintMixin):
    repr_attrs = ["name"]

    def __init__(
        self,
        name,
        metadata=None,
        config=None,
        reflect=False,
        skip_conversion_fields=False,
    ):
        self.name = DataSource.check_or_create_name(name)
        self._metrics = {}
        self._dimensions = {}
        self.graph = None

        config = config or {}
        if config:
            config = DataSourceConfigSchema().load(config)

        url = config.get("url", None)

        ds_config_context = zillion_config.get("DATASOURCE_CONTEXTS", {}).get(
            self.name, {}
        )
        if url and get_string_format_args(url):
            url = url.format(**ds_config_context)

        assert url != ADHOC_URL, "Unsupported datasource URL: '%s'" % url
        assert metadata or url, "You must pass metadata or config->url"
        assert not (
            url and metadata
        ), "Only one of metadata or config->url may be specified"

        if url:
            self.metadata = sa.MetaData()
            self.metadata.bind = sa.create_engine(url)
        else:
            assert isinstance(metadata, sa.MetaData), (
                "Invalid MetaData object: %s" % metadata
            )
            self.metadata = metadata
            assert (
                self.metadata.bind
            ), "MetaData object must have a bind (engine) attribute specified"

        self.reflect = reflect
        if reflect:
            self.reflect_metadata()

        self.apply_config(config, skip_conversion_fields=skip_conversion_fields)

    @classmethod
    def check_or_create_name(cls, name):
        if not name:
            datestr = datetime.datetime.utcnow().strftime("%Y%m%d%H%M%S")
            name = "zillion_ds_%s_%s" % (datestr, random.randint(0, 1e9))
            return name
        assert set(name) <= DATASOURCE_ALLOWABLE_CHARS, (
            'DataSource name "%s" has invalid characters. Allowed: %s'
            % (name, DATASOURCE_ALLOWABLE_CHARS_STR)
        )
        return name

    @classmethod
    def from_config(cls, name, config):
        return cls(name, reflect=True, config=config)

    @property
    def metric_tables(self):
        return {
            table_name: table
            for table_name, table in self.metadata.tables.items()
            if table.zillion and table.zillion.type == TableTypes.METRIC
        }

    @property
    def dimension_tables(self):
        return {
            table_name: table
            for table_name, table in self.metadata.tables.items()
            if table.zillion and table.zillion.type == TableTypes.DIMENSION
        }

    def get_dialect_name(self):
        return self.metadata.bind.dialect.name

    def reflect_metadata(self):
        dialect = self.get_dialect_name()
        schemas = get_schemas(self.metadata.bind)
        schemas = filter_dialect_schemas(schemas, dialect)
        for schema in schemas:
            self.metadata.reflect(schema=schema, views=True)

    def get_params(self):
        # TODO: does this need to store more information, entire config?
        return dict(
            name=self.name, url=str(self.metadata.bind.url), reflect=self.reflect
        )

    def print_info(self):
        print("---- Datasource %s" % self.name)
        print("metrics:")
        self.print_metrics(indent=2)
        print("dimensions:")
        self.print_dimensions(indent=2)

        print()
        for table in self.metadata.tables.values():
            print(format_msg("table: %s" % table.fullname, label=None))
            zillion_info = table.info.get("zillion", None)
            if not zillion_info:
                print(format_msg("table has no zillion info", label=None, indent=2))
                continue

            for column in table.c:
                print(format_msg("column: %s" % column.name, label=None, indent=2))
                zillion_info = column.info.get("zillion", None)
                if not zillion_info:
                    print(
                        format_msg("column has no zillion info", label=None, indent=4)
                    )
                    continue

                print(format_msg(column.info["zillion"], label=None, indent=4))

    def apply_table_configs(self, table_configs):
        """Take configs and apply them to the table/column metadata"""

        for table in self.metadata.tables.values():
            if table.fullname not in table_configs:
                continue

            table_config = table_configs[table.fullname]
            for param in ADHOC_TABLE_CONFIG_PARAMS:
                assert not table_config.get(param, None), (
                    "AdHoc table config param '%s' passed to non-adhoc datasource"
                    % param
                )

            table_info = TableInfo.schema_load(table_config, unknown=EXCLUDE)

            zillion_info = table.info.get("zillion", {})
            # Config takes precedence over values on table objects
            zillion_info.update(table_info)
            table.info["zillion"] = TableInfo.create(zillion_info)

            column_configs = table_config.get("columns", None)
            if not column_configs:
                continue

            for column in table.columns:
                if column.name not in column_configs:
                    continue

                column_config = column_configs[column.name]
                zillion_info = column.info.get("zillion", {})
                # Config takes precendence over values on column objects
                zillion_info.update(column_config)
                zillion_info["fields"] = zillion_info.get(
                    "fields", [default_field_name(column)]
                )
                column.info["zillion"] = ColumnInfo.create(zillion_info)

    def ensure_metadata_info(self):
        """Ensure that all zillion info are of proper type"""
        for table in self.metadata.tables.values():
            zillion_info = table.info.get("zillion", None)
            if not zillion_info:
                setattr(table, "zillion", None)
                continue

            table.info["zillion"] = TableInfo.create(zillion_info)
            setattr(table, "zillion", table.info["zillion"])

            column_count = 0

            for column in table.c:
                zillion_info = column.info.get("zillion", None) or {}
                if not zillion_info:
                    if not table.zillion.create_fields:
                        assert not column.primary_key, (
                            "Primary key column %s must have zillion info defined"
                            % column_fullname(column)
                        )
                        # If create_fields IS set the zillion info would
                        # automatically get created on the column and fields
                        # would automatically be created from the columns.
                        # Since it is NOT set, we just set the attribute to
                        # None and move on.
                        setattr(column, "zillion", None)
                        continue

                zillion_info["fields"] = zillion_info.get(
                    "fields", [default_field_name(column)]
                )
                if column.primary_key:
                    assert zillion_info["fields"], (
                        "Primary key column %s must have fields defined and one must be a valid dimension"
                        % column_fullname(column)
                    )
                column.info["zillion"] = ColumnInfo.create(zillion_info)
                setattr(column, "zillion", column.info["zillion"])
                column_count += 1

            assert column_count, (
                "Table %s has no columns with zillion info defined" % table.fullname
            )

    def add_conversion_fields(self):
        for table in self.metadata.tables.values():
            if not table.zillion:
                continue

            table_fields = get_table_fields(table)
            types_converted = set()

            for column in table.c:
                if not column.zillion:
                    continue

                if not column.zillion.allow_type_conversions:
                    # TODO: could allow specifying certain allowable conversions
                    # instead of just on/off switch
                    continue

                convs = get_dialect_type_conversions(self.get_dialect_name(), column)
                if convs:
                    assert not type(column.type) in types_converted, (
                        "Table %s has multiple columns of same type allowing conversions"
                        % table.fullname
                    )
                    types_converted.add(type(column.type))

                for field_name, ds_formula in convs:
                    if column.zillion.type_conversion_prefix:
                        field_name = column.zillion.type_conversion_prefix + field_name

                    if field_name in table_fields:
                        dbg(
                            "Skipping conversion field %s for column %s, already in table"
                            % (field_name, column_fullname(column))
                        )
                        continue
                    dbg(
                        "Adding conversion field %s for column %s"
                        % (field_name, column_fullname(column))
                    )
                    column.zillion.add_field(
                        dict(name=field_name, ds_formula=ds_formula)
                    )

    def add_metric_column(self, column, field):
        if not self.has_metric(field):
            dbg(
                "Adding metric %s from column %s.%s"
                % (field, self.name, column_fullname(column))
            )
            aggregation, rounding = infer_aggregation_and_rounding(column)
            metric = Metric(
                field, column.type, aggregation=aggregation, rounding=rounding
            )
            self.add_metric(metric)

    def add_dimension_column(self, column, field):
        if not self.has_dimension(field):
            dbg(
                "Adding dimension %s from column %s.%s"
                % (field, self.name, column_fullname(column))
            )
            dimension = Dimension(field, column.type)
            self.add_dimension(dimension)

    def add_metric_table_fields(self, table):
        for column in table.c:
            if not column.zillion:
                continue

            if not column.zillion.active:
                continue

            for field, field_def in column.zillion.get_fields().items():
                if self.has_field(field):
                    continue

                if not table.zillion.create_fields:
                    # If create_fields is False we do not automatically create fields
                    # from columns. The field would have to be explicitly defined
                    # in the metrics/dimensions of the datasource.
                    continue

                formula = (
                    field_def.get("ds_formula", None)
                    if isinstance(field_def, dict)
                    else None
                )
                if is_probably_metric(column, formula=formula):
                    self.add_metric_column(column, field)
                else:
                    self.add_dimension_column(column, field)

    def add_dimension_table_fields(self, table):
        for column in table.c:
            if not column.zillion:
                continue

            if not column.zillion.active:
                continue

            for field in column.zillion.get_field_names():
                if self.has_metric(field):
                    assert False, "Dimension table has metric field: %s" % field

                if self.has_dimension(field):
                    continue

                if not table.zillion.create_fields:
                    # If create_fields is False we do not automatically create fields
                    # from columns. The field would have to be explicitly defined
                    # in the metrics/dimensions of the datasource.
                    continue

                self.add_dimension_column(column, field)

    def populate_fields(self, config):
        self.populate_global_fields(config, force=True)

        for table in self.metadata.tables.values():
            if not table.zillion:
                continue
            if table.zillion.type == TableTypes.METRIC:
                self.add_metric_table_fields(table)
            elif table.zillion.type == TableTypes.DIMENSION:
                self.add_dimension_table_fields(table)
            else:
                assert False, "Invalid table type: %s" % table.zillion.type

    def get_primary_key_fields(self, primary_key):
        pk_fields = set()
        for col in primary_key:
            pk_dims = [
                x
                for x in col.zillion.fields
                if isinstance(x, str) and x in self.get_dimensions()
            ]
            # A primary key column can have multiple fields, but only one can
            # be a dimension so we clearly know what field to look for in
            # joins to related tables.
            assert len(pk_dims) == 1, (
                "Primary key column %s should have exactly one dimension, got: %s"
                % (col, pk_dims)
            )
            pk_fields.add(pk_dims[0])
        return pk_fields

    def find_neighbor_tables(self, table):
        neighbor_tables = []
        fields = get_table_fields(table)

        if table.zillion.type == TableTypes.METRIC:
            # Find dimension tables whose primary key is contained in the metric table
            for dim_table in self.dimension_tables.values():
                dt_pk_fields = self.get_primary_key_fields(dim_table.primary_key)
                can_join = True
                for field in dt_pk_fields:
                    if field not in fields:
                        can_join = False
                        break
                if can_join:
                    neighbor_tables.append(NeighborTable(dim_table, dt_pk_fields))

        # Add parent table if present
        parent_name = table.zillion.parent
        if parent_name:
            parent = self.metadata.tables[parent_name]
            pk_fields = self.get_primary_key_fields(parent.primary_key)
            for pk_field in pk_fields:
                assert pk_field in fields, (
                    "Table %s is parent of %s but primary key %s is not in both"
                    % (parent.fullname, table.fullname, pk_fields)
                )
            neighbor_tables.append(NeighborTable(parent, pk_fields))
        return neighbor_tables

    def build_ds_graph(self):
        graph = nx.DiGraph()
        self.graph = graph
        for table in self.metadata.tables.values():
            if not table.zillion:
                continue

            self.graph.add_node(table.fullname)
            neighbors = self.find_neighbor_tables(table)
            for neighbor in neighbors:
                self.graph.add_node(neighbor.table.fullname)
                self.graph.add_edge(
                    table.fullname,
                    neighbor.table.fullname,
                    join_fields=neighbor.join_fields,
                )

    def apply_config(self, config, skip_conversion_fields=False):
        if config.get("tables", None):
            self.apply_table_configs(config["tables"])

        self.ensure_metadata_info()

        if not skip_conversion_fields:
            self.add_conversion_fields()

        self.populate_fields(config)

        self.build_ds_graph()

    def has_table(self, table):
        return table.fullname in self.metadata.tables

    def get_table(self, fullname):
        return self.metadata.tables[fullname]

    def get_tables_with_field(self, field_name, table_type=None):
        tables = []
        for table in self.metadata.tables.values():
            if not table.zillion:
                continue
            if table_type and table.zillion.type != table_type:
                continue
            if field_name in get_table_fields(table):
                tables.append(table)
        return tables

    def get_metric_tables_with_metric(self, metric_name):
        return self.get_tables_with_field(metric_name, table_type=TableTypes.METRIC)

    def get_dim_tables_with_dim(self, dim_name):
        return self.get_tables_with_field(dim_name, table_type=TableTypes.DIMENSION)

    def get_columns_with_field(self, field_name):
        columns = []
        for table in self.metadata.tables.values():
            if not table.zillion:
                continue
            for col in table.c:
                if not (getattr(col, "zillion", None) and col.zillion.active):
                    continue
                if field_name in col.zillion.get_field_names():
                    columns.append(col)
        return columns

    def invert_field_joins(self, field_joins):
        """Take a map of fields to relevant joins and invert it"""
        join_fields = defaultdict(set)
        for field, joins in field_joins.items():
            for join in joins:
                if join in join_fields:
                    dbg("join %s already used, adding field %s" % (join, field))
                join_fields[join].add(field)
        return join_fields

    def populate_max_join_field_coverage(self, join_fields, grain):
        for join, covered_fields in join_fields.items():
            for field in grain:
                if field in covered_fields:
                    continue
                all_covered_fields = join.get_covered_fields()
                if field in all_covered_fields:
                    covered_fields.add(field)

    def eliminate_redundant_joins(self, sorted_join_fields):
        joins_to_delete = set()
        for join, covered_fields in sorted_join_fields:
            if join in joins_to_delete:
                continue

            dbg(
                "Finding redundant joins for %s / %s"
                % (join.table_names, covered_fields)
            )
            for other_join, other_covered_fields in sorted_join_fields:
                if join == other_join or join in joins_to_delete:
                    continue

                is_superset = join.table_names.issubset(other_join.table_names)
                has_unique_fields = other_covered_fields - covered_fields
                if is_superset and not has_unique_fields:
                    dbg(
                        "Removing redundant join %s / %s"
                        % (other_join.table_names, other_covered_fields)
                    )
                    joins_to_delete.add(other_join)

        sorted_join_fields = [
            (join, fields)
            for join, fields in sorted_join_fields
            if join not in joins_to_delete
        ]
        return sorted_join_fields

    def find_join_combinations(self, sorted_join_fields, grain):
        candidates = []
        for join_combo in powerset(sorted_join_fields):
            if not join_combo:
                continue

            covered = set()
            has_subsets = False
            for join, covered_dims in join_combo:
                covered |= covered_dims
                # If any of the other joins are a subset of this join we ignore it.
                # The powerset has every combination so it will eventually hit the case
                # where there are only distinct joins.
                for other_join, other_covered_dims in join_combo:
                    if join == other_join:
                        continue
                    if other_join.table_names.issubset(join.table_names):
                        has_subsets = True
                        break
                if has_subsets:
                    break

            if has_subsets:
                continue

            if len(covered) == len(grain):
                # This combination of joins covers the entire grain. Add it as a candidate if
                # there isn't an existing candidate that is a a subset of these joins
                skip = False
                joins = set([x[0] for x in join_combo])
                for other_join_combo in candidates:
                    other_joins = set([x[0] for x in other_join_combo])
                    if other_joins.issubset(joins):
                        skip = True
                if skip:
                    dbg("Skipping subset join list combination")
                    continue
                candidates.append(join_combo)

        return candidates

    def choose_best_join_combination(self, candidates):
        ordered = sorted(
            candidates, key=lambda x: len(iter_or([y[0].table_names for y in x]))
        )
        chosen = ordered[0]
        join_fields = {}
        for join, covered_fields in chosen:
            join_fields[join] = covered_fields
            join.add_fields(covered_fields)
        return join_fields

    def consolidate_field_joins(self, grain, field_joins):
        """This takes a mapping of fields to joins that satisfy that field
        and returns a minimized map of joins to fields satisfied by that join.
        """

        # Some preliminary shuffling of the inputs to support later logic
        join_fields = self.invert_field_joins(field_joins)
        self.populate_max_join_field_coverage(join_fields, grain)

        # Sort by number of dims covered desc, number of tables involved asc
        sorted_join_fields = sorted(
            join_fields.items(), key=lambda kv: (len(kv[1]), -len(kv[0])), reverse=True
        )

        if len(sorted_join_fields[0][1]) == len(grain):
            # Single join covers entire grain. It should be ~optimal based on sorting.
            join = sorted_join_fields[0][0]
            covered_fields = sorted_join_fields[0][1]
            join.add_fields(covered_fields)
            return {join: covered_fields}

        sorted_join_fields = self.eliminate_redundant_joins(sorted_join_fields)
        candidates = self.find_join_combinations(sorted_join_fields, grain)
        join_fields = self.choose_best_join_combination(candidates)
        return join_fields

    def find_joins_to_dimension(self, table, dimension):
        joins = []

        dim_columns = self.get_columns_with_field(dimension)
        dim_column_table_map = {c.table.fullname: c for c in dim_columns}

        for column in dim_columns:
            if column.table == table:
                paths = [[table.fullname]]
            else:
                # TODO: consider caching with larger graphs, or precomputing
                paths = nx.all_simple_paths(
                    self.graph, table.fullname, column.table.fullname
                )

            if not paths:
                continue

            for path in paths:
                # For each path, if this dim can be found earlier in the path then
                # reference it in the earlier (child) table
                field_map = None
                for table_name in path:
                    if table_name in dim_column_table_map:
                        field_map = {dimension: dim_column_table_map[table_name]}
                        break

                assert field_map, "Could not map dimension %s to column" % dimension
                join = joins_from_path(self, path, field_map=field_map)
                joins.append(join)

        dbg("Found joins to dim %s for table %s:" % (dimension, table.fullname))
        dbg(joins)
        return joins

    def get_possible_joins(self, table, grain):
        """This takes a given table (usually a metric table) and tries to find one or
        more joins to each dimension of the grain. It's possible some of these
        joins satisfy other parts of the grain too which leaves room for
        consolidation, but it's also possible to have it generate independent,
        non-overlapping joins to meet the grain.
        """
        assert self.has_table(table), "Could not find table %s in datasource %s" % (
            table.fullname,
            self.name,
        )

        if not grain:
            dbg("No grain specified, ignoring joins")
            return None

        possible_dim_joins = {}
        for dimension in grain:
            dim_joins = self.find_joins_to_dimension(table, dimension)
            if not dim_joins:
                dbg(
                    "table %s can not satisfy dimension %s"
                    % (table.fullname, dimension)
                )
                return None

            possible_dim_joins[dimension] = dim_joins

        possible_joins = self.consolidate_field_joins(grain, possible_dim_joins)
        dbg("possible joins:")
        dbg(possible_joins)
        return possible_joins

    def find_possible_table_sets(self, ds_tables_with_field, field, grain):
        table_sets = []
        for field_table in ds_tables_with_field:
            if (not grain) or grain.issubset(get_table_fields(field_table)):
                table_set = TableSet(self, field_table, None, grain, set([field]))
                table_sets.append(table_set)
                dbg("full grain (%s) covered in %s" % (grain, field_table.fullname))
                continue

            joins = self.get_possible_joins(field_table, grain)
            if not joins:
                dbg("table %s can not join at grain %s" % (field_table.fullname, grain))
                continue

            dbg(
                "adding %d possible join(s) to table %s"
                % (len(joins), field_table.fullname)
            )
            for join, covered_dims in joins.items():
                table_set = TableSet(self, field_table, join, grain, set([field]))
                table_sets.append(table_set)

        return table_sets


class AdHocDataTable(PrintMixin):
    repr_attrs = ["name", "primary_key", "table_config"]

    @initializer
    def __init__(
        self,
        name,
        data,
        table_type,
        columns=None,
        primary_key=None,
        parent=None,
        schema=None,
        **kwargs
    ):
        self.df_kwargs = kwargs or {}
        self.table_config = TableConfigSchema().load(
            dict(
                type=table_type, columns=self.columns, create_fields=True, parent=parent
            )
        )

    @property
    def fullname(self):
        if self.schema:
            return "%s.%s" % (self.schema, self.name)
        return self.name

    def get_dataframe(self):
        if isinstance(self.data, pd.DataFrame):
            return self.data

        kwargs = self.df_kwargs.copy()
        if self.columns:
            kwargs["columns"] = self.columns
        return pd.DataFrame.from_records(self.data, self.primary_key, **kwargs)

    def table_exists(self, engine):
        qr = engine.execute(
            "SELECT name FROM sqlite_master " "WHERE type='table' AND name=?",
            (self.name),
        )
        result = qr.fetchone()
        if result:
            return True
        return False

    def to_sql(self, engine, if_exists="fail", method="multi", chunksize=int(1e3)):
        if if_exists == "ignore":
            if self.table_exists(engine):
                return
            # Pandas doesn't actually have an "ignore" option, but switching
            # to fail will work because the table *should* not exist.
            if_exists = "fail"

        df = self.get_dataframe()

        # Note: this hits limits in allowed sqlite params if chunks are too large
        df.to_sql(
            self.name,
            engine,
            if_exists=if_exists,
            method=method,
            chunksize=chunksize,
            schema=self.schema,
        )


class SQLiteDataTable(AdHocDataTable):
    def get_dataframe(self):
        raise NotImplementedError

    def to_sql(self, engine, **kwargs):
        assert self.table_exists(engine), "SQLiteDataTable table does not exist"


class CSVDataTable(AdHocDataTable):
    def get_dataframe(self):
        return pd.read_csv(
            self.data,
            index_col=self.primary_key,
            usecols=list(self.columns.keys()) if self.columns else None,
            **self.df_kwargs
        )


class ExcelDataTable(AdHocDataTable):
    def get_dataframe(self):
        df = pd.read_excel(
            self.data,
            usecols=list(self.columns.keys()) if self.columns else None,
            **self.df_kwargs
        )
        if self.primary_key and df.index.names != self.primary_key:
            df = df.set_index(self.primary_key)
        return df


class JSONDataTable(AdHocDataTable):
    def get_dataframe(self, orient="table"):
        df = pd.read_json(self.data, orient=orient, **self.df_kwargs)
        if self.primary_key and df.index.names != self.primary_key:
            df = df.set_index(self.primary_key)
        return df


class HTMLDataTable(AdHocDataTable):
    def get_dataframe(self):
        # Expects this format by default:
        # df.reset_index().to_html("dma_zip.html", index=False)
        dfs = pd.read_html(self.data, **self.df_kwargs)
        assert dfs, "No html table found"
        assert len(dfs) == 1, "More than one html table found"
        df = dfs[0]
        if self.primary_key and df.index.names != self.primary_key:
            df = df.set_index(self.primary_key)
        return df


class GoogleSheetsDataTable(AdHocDataTable):
    def get_dataframe(self):
        parsed = urlparse(self.data)
        params = parse_qs(parsed.query)
        if params.get("format", None) == ["csv"]:
            cls = CSVDataTable
        elif parsed.path.endswith("/edit"):
            parsed = parsed._replace(
                path=parsed.path.replace("/edit", "/export"), query="format=csv"
            )
            url = urlunparse(parsed)
        else:
            raise Exception("Unsupported google docs URL: %s" % url)

        return pd.read_csv(
            url,
            index_col=self.primary_key,
            usecols=list(self.columns.keys()) if self.columns else None,
            **self.df_kwargs
        )


class AdHocDataSource(DataSource):
    def __init__(
        self, datatables, name=None, config=None, coerce_float=True, if_exists="fail"
    ):
        config = config or dict(tables={})
        ds_name = self.check_or_create_name(name)

        conn_url = self.get_datasource_url(ds_name)
        engine = sa.create_engine(conn_url, echo=False)

        for dt in datatables:
            dt.to_sql(engine, if_exists=if_exists)
            config.setdefault("tables", {})[dt.fullname] = dt.table_config

        metadata = sa.MetaData()
        metadata.bind = engine

        super(AdHocDataSource, self).__init__(
            ds_name, metadata=metadata, config=config, reflect=True
        )

    @classmethod
    def from_config(cls, name, config, if_exists="fail"):
        for table_name, table_config in config["tables"].items():
            assert table_config.get(
                "url", None
            ), "All tables in an adhoc datasource config must have a url"

        ds_config_context = zillion_config.get("DATASOURCE_CONTEXTS", {}).get(name, {})

        datatables = []
        for table_name, table_config in config["tables"].items():
            cfg = table_config.copy()
            schema = None

            if get_string_format_args(cfg["url"]):
                cfg["url"] = cfg["url"].format(**ds_config_context)

            if "." in table_name:
                parts = table_name.split(".")
                # This is also checked in config, should never happen
                assert len(parts) == 2, "Invalid table name: %s" % table_name
                schema, table_name = parts

            dt = datatable_from_config(table_name, cfg, schema=schema)
            datatables.append(dt)

        return cls(datatables, name=name, if_exists=if_exists)

    @classmethod
    def get_datasource_filename(cls, ds_name):
        dir_name = zillion_config["ADHOC_DATASOURCE_DIRECTORY"]
        return "%s/%s.db" % (dir_name, ds_name)

    @classmethod
    def get_datasource_url(cls, ds_name):
        return "sqlite:///%s" % cls.get_datasource_filename(ds_name)

    def clean_up(self):
        filename = self.get_datasource_filename(self.name)
        dbg("Removing %s" % filename)
        rmfile(filename)


def datasource_from_config(name, config, if_exists="fail"):
    if config.get("url", None) == ADHOC_URL:
        return AdHocDataSource.from_config(name, config, if_exists=if_exists)
    return DataSource.from_config(name, config)


def datatable_from_config(name, config, schema=None, **kwargs):
    assert config.get(
        "create_fields", True
    ), "AdHocDataTables must have create_fields=True"

    url = config["url"]
    if url.endswith("csv"):
        cls = CSVDataTable
    elif url.endswith("xlsx") or url.endswith("xls"):
        cls = ExcelDataTable
    elif url.endswith("json"):
        cls = JSONDataTable
    elif url.endswith("html"):
        cls = HTMLDataTable
    elif "docs.google.com" in url:
        cls = GoogleSheetsDataTable

    kwargs.update(config.get("adhoc_table_options", {}))

    return cls(
        name,
        url,
        config["type"],
        config.get("columns", None),
        primary_key=config.get("primary_key", None),
        parent=config.get("parent", None),
        schema=schema,
        **kwargs
    )



from collections import OrderedDict
import os
import string

from marshmallow import (
    Schema,
    fields as mfields,
    ValidationError,
    validates_schema,
    pre_load,
    EXCLUDE,
    INCLUDE,
    RAISE,
)
from pandas.io.common import get_filepath_or_buffer
from tlbx import dbg, error, json, st, initializer, MappingMixin, PrintMixin
import yaml

from zillion.core import FieldTypes, TableTypes, AggregationTypes, TechnicalTypes
from zillion.sql_utils import (
    column_fullname,
    type_string_to_sa_type,
    InvalidSQLAlchemyTypeString,
)


FIELD_ALLOWABLE_CHARS_STR = (
    string.ascii_uppercase + string.ascii_lowercase + string.digits + "_"
)
FIELD_ALLOWABLE_CHARS = set(FIELD_ALLOWABLE_CHARS_STR)

DATASOURCE_ALLOWABLE_CHARS_STR = (
    string.ascii_uppercase + string.ascii_lowercase + string.digits + "_"
)
DATASOURCE_ALLOWABLE_CHARS = set(DATASOURCE_ALLOWABLE_CHARS_STR)


def field_safe_name(name):
    for char in name:
        if char not in FIELD_ALLOWABLE_CHARS:
            name = name.replace(char, "_")
    return name


def default_field_name(column):
    return field_safe_name(column_fullname(column))


def parse_technical_string(val):
    min_period = None
    center = None
    parts = val.split("-")

    if len(parts) == 2:
        ttype, window = parts
    elif len(parts) == 3:
        ttype, window, min_period = parts

    val = dict(type=ttype, window=window)
    if min_period is not None:
        val["min_period"] = min_period
    if center is not None:
        val["center"] = center
    return val


def load_zillion_config():
    zillion_config_fname = os.environ.get("ZILLION_CONFIG", None)
    if not zillion_config_fname:
        return dict(
            DEBUG=False,
            ZILLION_DB_URL="sqlite:////tmp/zillion.db",
            ADHOC_DATASOURCE_DIRECTORY="/tmp",
            LOAD_TABLE_CHUNK_SIZE=5000,
            IFNULL_PRETTY_VALUE="--",
            DATASOURCE_QUERY_MODE="sequential",
            DATASOURCE_QUERY_TIMEOUT=None,
            DATASOURCE_CONTEXTS={},
        )
    return yaml.safe_load(open(zillion_config_fname))


zillion_config = load_zillion_config()


def parse_schema_file(filename, schema, object_pairs_hook=None):
    """Parse a marshmallow schema file"""
    f = open(filename)
    raw = f.read()
    f.close()
    try:
        # This does the schema check, but has a bug in object_pairs_hook so order is not preserved
        if object_pairs_hook:
            assert False, "Needs to support marshmallow pre_load behavior somehow"
            result = json.loads(raw, object_pairs_hook=object_pairs_hook)
        else:
            result = schema.loads(raw)
    except ValidationError as e:
        error("Schema Validation Error: %s" % schema)
        print(json.dumps(str(e), indent=2))
        raise
    return result


def load_warehouse_config(filename, preserve_order=False):
    file_schema = WarehouseConfigSchema()
    config = parse_schema_file(
        filename, file_schema, object_pairs_hook=OrderedDict if preserve_order else None
    )
    return config


def load_datasource_config(filename, preserve_order=False):
    file_schema = DataSourceConfigSchema()
    config = parse_schema_file(
        filename, file_schema, object_pairs_hook=OrderedDict if preserve_order else None
    )
    return config


def is_valid_table_type(val):
    if val in TableTypes:
        return True
    raise ValidationError("Invalid table type: %s" % val)


def is_valid_table_name(val):
    if val.count(".") > 1:
        raise ValidationError("Table name has more than one period: %s" % val)
    return True


def is_valid_field_name(val):
    if val is None:
        raise ValidationError("Field name can not be null")
    if set(val) <= FIELD_ALLOWABLE_CHARS:
        return True
    raise ValidationError(
        'Field name "%s" has invalid characters. Allowed: %s'
        % (val, FIELD_ALLOWABLE_CHARS_STR)
    )


def is_valid_sqlalchemy_type(val):
    if val is not None:
        try:
            sa_type = type_string_to_sa_type(val)
        except InvalidSQLAlchemyTypeString:
            raise ValidationError("Invalid table type: %s" % val)
    return True


def is_valid_aggregation(val):
    if val in AggregationTypes:
        return True
    raise ValidationError("Invalid aggregation: %s" % val)


def is_valid_column_field_config(val):
    if isinstance(val, str):
        return True
    if isinstance(val, dict):
        schema = ColumnFieldConfigSchema()
        schema.load(val)
        return True
    raise ValidationError("Invalid column field config: %s" % val)


def is_valid_technical_type(val):
    if val in TechnicalTypes:
        return True
    raise ValidationError("Invalid technical type: %s" % val)


def is_valid_technical(val):
    if isinstance(val, str):
        val = parse_technical_string(val)
    elif not isinstance(val, dict):
        raise ValidationError("Invalid technical: %s" % val)
    schema = TechnicalInfoSchema()
    val = schema.load(val)
    return True


def is_valid_datasource_config(val):
    if not isinstance(val, dict):
        raise ValidationError("Invalid datasource config: %s" % val)
    schema = DataSourceConfigSchema()
    val = schema.load(val)
    return True


class BaseSchema(Schema):
    class Meta:
        # Use the json module as imported from tlbx
        json_module = json


class TechnicalInfoSchema(BaseSchema):
    type = mfields.String(required=True, validate=is_valid_technical_type)
    window = mfields.Integer(required=True)
    min_periods = mfields.Integer(default=1, missing=1)
    center = mfields.Boolean(default=False, missing=False)


class TechnicalField(mfields.Field):
    def _validate(self, value):
        is_valid_technical(value)
        super()._validate(value)


class AdHocFieldSchema(BaseSchema):
    name = mfields.String(required=True, validate=is_valid_field_name)
    formula = mfields.String(required=True)


class AdHocMetricSchema(AdHocFieldSchema):
    technical = TechnicalField(default=None, missing=None)
    rounding = mfields.Integer(default=None, missing=None)


class ColumnFieldConfigSchema(BaseSchema):
    name = mfields.Str(required=True, validate=is_valid_field_name)
    ds_formula = mfields.Str(required=True)


class ColumnFieldConfigField(mfields.Field):
    def _validate(self, value):
        is_valid_column_field_config(value)
        super()._validate(value)


class ColumnInfoSchema(BaseSchema):
    fields = mfields.List(ColumnFieldConfigField())
    allow_type_conversions = mfields.Boolean(default=False, missing=False)
    type_conversion_prefix = mfields.String(default=None, missing=None)
    active = mfields.Boolean(default=True, missing=True)


class ColumnConfigSchema(ColumnInfoSchema):
    pass


class TableTypeField(mfields.Field):
    def _validate(self, value):
        is_valid_table_type(value)
        super()._validate(value)


class TableInfoSchema(BaseSchema):
    type = TableTypeField(required=True)
    active = mfields.Boolean(default=True, missing=True)
    parent = mfields.Str(default=None, missing=None)
    create_fields = mfields.Boolean(default=False, missing=False)


ADHOC_TABLE_CONFIG_PARAMS = ["url", "primary_key", "adhoc_table_options"]


class TableConfigSchema(TableInfoSchema):
    columns = mfields.Dict(
        keys=mfields.Str(),
        values=mfields.Nested(ColumnConfigSchema),
        missing=None,
        required=False,
    )
    url = mfields.String()
    primary_key = mfields.List(mfields.String())
    adhoc_table_options = mfields.Dict(keys=mfields.Str())


class MetricConfigSchema(BaseSchema):
    name = mfields.String(required=True, validate=is_valid_field_name)
    type = mfields.String(default=None, missing=None, validate=is_valid_sqlalchemy_type)
    aggregation = mfields.String(
        default=AggregationTypes.SUM,
        missing=AggregationTypes.SUM,
        validate=is_valid_aggregation,
    )
    rounding = mfields.Integer(default=None, missing=None)
    weighting_metric = mfields.Str(default=None, missing=None)
    formula = mfields.String(default=None, missing=None)
    technical = TechnicalField(default=None, missing=None)

    @validates_schema(skip_on_field_errors=True)
    def validate_object(self, data, **kwargs):
        if (not data.get("type", None)) and (not data.get("formula", None)):
            raise ValidationError(
                "Either type or formula must be specified for metric: %s" % data
            )

        if data["weighting_metric"] and not data["aggregation"] == AggregationTypes.AVG:
            raise ValidationError(
                'only "%s" aggregation type is allowed with weighting metrics: %s'
                % (AggregationTypes.AVG, data)
            )


class DimensionConfigSchema(BaseSchema):
    name = mfields.String(required=True, validate=is_valid_field_name)
    type = mfields.String(default=None, missing=None, validate=is_valid_sqlalchemy_type)
    # TODO: add support for FormulaDimensions
    # formula = mfields.String(default=None, missing=None)


class TableNameField(mfields.Str):
    def _validate(self, value):
        is_valid_table_name(value)
        super()._validate(value)


class DataSourceConfigSchema(BaseSchema):
    url = mfields.String()
    metrics = mfields.List(mfields.Nested(MetricConfigSchema))
    dimensions = mfields.List(mfields.Nested(DimensionConfigSchema))
    tables = mfields.Dict(
        keys=TableNameField(), values=mfields.Nested(TableConfigSchema)
    )

    @pre_load
    def check_table_refs(self, data, **kwargs):
        for table_name, table_config in data.get("tables", {}).items():
            if not isinstance(table_config, str):
                continue

            f, _, _, should_close = get_filepath_or_buffer(table_config)
            close = False or should_close
            if isinstance(f, str):
                f = open(f, "r")
                close = True

            try:
                raw = f.read()
            finally:
                if close:
                    try:
                        f.close()
                    except ValueError:
                        pass

            json_config = json.loads(raw)
            schema = TableConfigSchema()
            config = schema.load(json_config)
            data["tables"][table_name] = config

        return data


class DataSourceConfigField(mfields.Field):
    def _validate(self, value):
        is_valid_datasource_config(value)
        super()._validate(value)


class WarehouseConfigSchema(BaseSchema):
    metrics = mfields.List(mfields.Nested(MetricConfigSchema))
    dimensions = mfields.List(mfields.Nested(DimensionConfigSchema))
    datasources = mfields.Dict(
        keys=mfields.Str(), values=DataSourceConfigField, required=True
    )

    @pre_load
    def check_ds_refs(self, data, **kwargs):
        for ds_name, ds_config in data.get("datasources", {}).items():
            if not isinstance(ds_config, str):
                continue

            f, _, _, should_close = get_filepath_or_buffer(ds_config)
            close = False or should_close
            if isinstance(f, str):
                f = open(f, "r")
                close = True

            try:
                raw = f.read()
            finally:
                if close:
                    try:
                        f.close()
                    except ValueError:
                        pass

            json_config = json.loads(raw)
            schema = DataSourceConfigSchema()
            config = schema.load(json_config)
            data["datasources"][ds_name] = config

        return data


class ZillionInfo(MappingMixin):
    schema = None

    @initializer
    def __init__(self, **kwargs):
        assert self.schema, "ZillionInfo subclass must have a schema defined"
        self.schema().load(self)

    @classmethod
    def schema_validate(cls, zillion_info, unknown=RAISE):
        return cls.schema(unknown=unknown).validate(zillion_info)

    @classmethod
    def schema_load(cls, zillion_info, unknown=RAISE):
        return cls.schema().load(zillion_info, unknown=unknown)

    @classmethod
    def create(cls, zillion_info, unknown=RAISE):
        if isinstance(zillion_info, cls):
            return zillion_info
        assert isinstance(zillion_info, dict), (
            "Raw info must be a dict: %s" % zillion_info
        )
        zillion_info = cls.schema().load(zillion_info, unknown=unknown)
        return cls(**zillion_info)


class TableInfo(ZillionInfo, PrintMixin):
    repr_attrs = ["type", "active", "create_fields", "parent"]
    schema = TableInfoSchema


class ColumnInfo(ZillionInfo, PrintMixin):
    repr_attrs = ["fields", "active"]
    schema = ColumnInfoSchema

    def __init__(self, **kwargs):
        super(ColumnInfo, self).__init__(**kwargs)
        self.field_map = OrderedDict()
        for field in self.fields:
            self.add_field_to_map(field)

    def has_field(self, field):
        if not isinstance(field, str):
            field = field["name"]
        if field in self.field_map:
            return True
        return False

    def add_field_to_map(self, field):
        assert not self.has_field(field), "Field %s is already added" % field
        if isinstance(field, str):
            self.field_map[field] = None
        else:
            # TODO: FieldInfoSchema?
            assert isinstance(field, dict) and "name" in field, (
                "Invalid field config: %s" % field
            )
            self.field_map[field["name"]] = field

    def add_field(self, field):
        self.add_field_to_map(field)
        self.fields.append(field)

    def get_fields(self):
        return {k: v for k, v in self.field_map.items()}

    def get_field_names(self):
        return self.field_map.keys()



import sqlalchemy as sa
from tlbx import (
    MappingMixin,
    PrintMixin,
    initializer,
    warn,
    info,
    format_msg,
    st,
    get_string_format_args,
)
from zillion.configs import (
    MetricConfigSchema,
    TechnicalInfoSchema,
    DimensionConfigSchema,
    AdHocMetricSchema,
    AdHocFieldSchema,
    parse_technical_string,
    is_valid_field_name,
    zillion_config,
)
from zillion.core import (
    InvalidFieldException,
    MaxFormulaDepthException,
    DisallowedSQLException,
    AggregationTypes,
    TableTypes,
    FieldTypes,
)
from zillion.sql_utils import (
    aggregation_to_sqla_func,
    contains_aggregation,
    contains_sql_keywords,
    type_string_to_sa_type,
    is_probably_metric,
    sqla_compile,
    column_fullname,
)


MAX_FORMULA_DEPTH = 3


class Technical(MappingMixin, PrintMixin):
    repr_attrs = ["type", "window", "min_periods"]

    @initializer
    def __init__(self, **kwargs):
        # Attributes of this will be used to apply DataFrame.rolling
        # to result columns
        pass

    @classmethod
    def create(cls, info):
        if isinstance(info, cls):
            return info
        if isinstance(info, str):
            info = parse_technical_string(info)
        assert isinstance(info, dict), "Raw info must be a dict: %s" % info
        info = TechnicalInfoSchema().load(info)
        return cls(**info)


class Field(PrintMixin):
    repr_attrs = ["name"]
    field_type = None

    @initializer
    def __init__(self, name, type, **kwargs):
        is_valid_field_name(name)
        if isinstance(type, str):
            self.type = type_string_to_sa_type(type)

    def get_formula_fields(self, warehouse, depth=0, adhoc_fms=None):
        return None, None

    def get_ds_expression(self, column, label=True):
        ds_formula = (
            column.zillion.field_map[self.name].get("ds_formula", None)
            if column.zillion.field_map[self.name]
            else None
        )
        if not ds_formula:
            if label:
                return column.label(self.name)
            return column

        if contains_sql_keywords(ds_formula):
            raise DisallowedSQLException(
                "Formula contains disallowed sql: %s" % ds_formula
            )

        if not (ds_formula.startswith("(") and ds_formula.endswith("(")):
            ds_formula = "(" + ds_formula + ")"

        if label:
            return sa.literal_column(ds_formula).label(self.name)
        return sa.literal_column(ds_formula)

    def get_final_select_clause(self, *args, **kwargs):
        return self.name

    # https://stackoverflow.com/questions/2909106/whats-a-correct-and-good-way-to-implement-hash
    def __key(self):
        return self.name

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(self, type(other)) and self.__key() == other.__key()


class Metric(Field):
    field_type = FieldTypes.METRIC

    def __init__(
        self,
        name,
        type,
        aggregation=AggregationTypes.SUM,
        rounding=None,
        weighting_metric=None,
        technical=None,
        **kwargs
    ):
        if weighting_metric:
            assert aggregation == AggregationTypes.AVG, (
                'Weighting metrics are only supported for "%s" aggregation type'
                % AggregationTypes.AVG
            )

        if technical:
            technical = Technical.create(technical)

        super(Metric, self).__init__(
            name,
            type,
            aggregation=aggregation,
            rounding=rounding,
            weighting_metric=weighting_metric,
            technical=technical,
            **kwargs
        )

    def get_ds_expression(self, column, label=True):
        expr = column
        aggr = aggregation_to_sqla_func(self.aggregation)
        skip_aggr = False

        ds_formula = (
            column.zillion.field_map[self.name].get("ds_formula", None)
            if column.zillion.field_map[self.name]
            else None
        )

        if ds_formula:
            if contains_sql_keywords(ds_formula):
                raise DisallowedSQLException(
                    "Formula contains disallowed sql: %s" % ds_formula
                )
            if contains_aggregation(ds_formula):
                info("Datasource formula contains aggregation, skipping default logic")
                skip_aggr = True
            expr = sa.literal_column(ds_formula)

        if not skip_aggr:
            if self.aggregation in [
                AggregationTypes.COUNT,
                AggregationTypes.COUNT_DISTINCT,
            ]:
                if self.rounding:
                    info("Ignoring rounding for count field: %s" % self.name)
                if label:
                    return aggr(expr).label(self.name)
                return aggr(expr)

            if self.weighting_metric:
                w_column = get_table_field_column(column.table, self.weighting_metric)
                w_column_name = column_fullname(w_column)
                # NOTE: 1.0 multiplication is a hack to ensure results are not rounded
                # to integer values improperly by some database dialects such as sqlite
                expr = sa.func.sum(
                    sa.text("1.0") * expr * sa.text(w_column_name)
                ) / sa.func.sum(sa.text(w_column_name))
            else:
                expr = aggr(expr)

        if label:
            return expr.label(self.name)
        return expr

    def get_final_select_clause(self, *args, **kwargs):
        return self.name


class Dimension(Field):
    field_type = FieldTypes.DIMENSION


class FormulaField(Field):
    def __init__(self, name, formula, **kwargs):
        super(FormulaField, self).__init__(name, None, formula=formula, **kwargs)

    def get_formula_fields(self, warehouse, depth=0, adhoc_fms=None):
        if depth > MAX_FORMULA_DEPTH:
            raise MaxFormulaDepthException

        raw_formula = self.formula
        raw_fields = set()
        formula_fields = get_string_format_args(self.formula)
        field_formula_map = {}

        for field_name in formula_fields:
            field = warehouse.get_field(field_name, adhoc_fms=adhoc_fms)
            if isinstance(field, FormulaField):
                try:
                    sub_fields, sub_formula = field.get_formula_fields(
                        warehouse, depth=depth + 1, adhoc_fms=adhoc_fms
                    )
                except MaxFormulaDepthException:
                    if depth != 0:
                        raise
                    raise MaxFormulaDepthException(
                        "Maximum formula recursion depth exceeded for %s: %s"
                        % (self.name, self.formula)
                    )
                for sub_field in sub_fields:
                    raw_fields.add(sub_field)
                field_formula_map[field_name] = "(" + sub_formula + ")"
            else:
                field_formula_map[field_name] = "{" + field_name + "}"
                raw_fields.add(field_name)

        raw_formula = self.formula.format(**field_formula_map)
        return raw_fields, raw_formula

    def check_formula_fields(self, warehouse, adhoc_fms=None):
        fields, _ = self.get_formula_fields(warehouse, adhoc_fms=adhoc_fms)
        for field in fields:
            warehouse.get_field(field, adhoc_fms=adhoc_fms)

    def get_ds_expression(self, column, label=True):
        assert False, "Formula-based Fields do not support get_ds_expression"

    def get_final_select_clause(self, warehouse, adhoc_fms=None):
        formula_fields, raw_formula = self.get_formula_fields(
            warehouse, adhoc_fms=adhoc_fms
        )
        format_args = {k: k for k in formula_fields}
        formula = raw_formula.format(**format_args)
        if contains_sql_keywords(formula):
            raise DisallowedSQLException(
                "Formula contains disallowed sql: %s" % formula
            )
        return sqla_compile(sa.text(formula))


class FormulaDimension(FormulaField):
    repr_atts = ["name", "formula"]
    field_type = FieldTypes.DIMENSION

    def __init__(self, name, formula, **kwargs):
        # super(FormulaDimension, self).__init__(
        #     name,
        #     formula,
        #     **kwargs
        # )
        raise InvalidFieldException("FormulaDimensions are not currently supported")


class FormulaMetric(FormulaField):
    repr_atts = ["name", "formula", "technical"]
    field_type = FieldTypes.METRIC

    def __init__(
        self,
        name,
        formula,
        aggregation=AggregationTypes.SUM,
        rounding=None,
        weighting_metric=None,
        technical=None,
        **kwargs
    ):
        if technical:
            technical = Technical.create(technical)

        super(FormulaMetric, self).__init__(
            name,
            formula,
            aggregation=aggregation,
            rounding=rounding,
            weighting_metric=weighting_metric,
            technical=technical,
            **kwargs
        )


class AdHocField(FormulaField):
    @classmethod
    def create(cls, obj):
        schema = AdHocFieldSchema()
        field_def = schema.load(obj)
        return cls(field_def["name"], field_def["formula"])


class AdHocMetric(FormulaMetric):
    def __init__(self, name, formula, technical=None, rounding=None):
        super(AdHocMetric, self).__init__(
            name, formula, technical=technical, rounding=rounding
        )

    @classmethod
    def create(cls, obj):
        schema = AdHocMetricSchema()
        field_def = schema.load(obj)
        return cls(
            field_def["name"],
            field_def["formula"],
            technical=field_def["technical"],
            rounding=field_def["rounding"],
        )


class AdHocDimension(AdHocField):
    field_type = FieldTypes.DIMENSION


def create_metric(metric_def):
    if metric_def["formula"]:
        metric = FormulaMetric(
            metric_def["name"],
            metric_def["formula"],
            aggregation=metric_def["aggregation"],
            rounding=metric_def["rounding"],
            weighting_metric=metric_def["weighting_metric"],
            technical=metric_def["technical"],
        )
    else:
        metric = Metric(
            metric_def["name"],
            metric_def["type"],
            aggregation=metric_def["aggregation"],
            rounding=metric_def["rounding"],
            weighting_metric=metric_def["weighting_metric"],
            technical=metric_def["technical"],
        )
    return metric


def create_dimension(dim_def):
    if dim_def.get("formula", None):
        # dim = FormulaDimension(dim_def["name"], dim_def["formula"])
        raise InvalidFieldException("FormulaDimensions are not currently supported")
    else:
        dim = Dimension(dim_def["name"], dim_def["type"])
    return dim


class FieldManagerMixin:
    metrics_attr = "_metrics"
    dimensions_attr = "_dimensions"

    def get_child_field_managers(self):
        return []

    def get_field_managers(self, adhoc_fms=None):
        return self.get_child_field_managers() + (adhoc_fms or [])

    def _directly_has_metric(self, name):
        return name in getattr(self, self.metrics_attr)

    def _directly_has_dimension(self, name):
        return name in getattr(self, self.dimensions_attr)

    def _directly_has_field(self, name):
        return name in getattr(self, self.metrics_attr) or name in getattr(
            self, self.dimensions_attr
        )

    def print_metrics(self, indent=None):
        print(format_msg(getattr(self, self.metrics_attr), label=None, indent=indent))

    def print_dimensions(self, indent=None):
        print(
            format_msg(getattr(self, self.dimensions_attr), label=None, indent=indent)
        )

    def has_metric(self, name, adhoc_fms=None):
        if self._directly_has_metric(name):
            return True
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            if fm.has_metric(name):
                return True
        return False

    def has_dimension(self, name, adhoc_fms=None):
        if self._directly_has_dimension(name):
            return True
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            if fm.has_dimension(name):
                return True
        return False

    def has_field(self, name, adhoc_fms=None):
        if self._directly_has_field(name):
            return True
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            if fm.has_field(name):
                return True
        return False

    def get_metric(self, obj, adhoc_fms=None):
        if isinstance(obj, str):
            if self._directly_has_metric(obj):
                return getattr(self, self.metrics_attr)[obj]
            for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
                if fm.has_metric(obj):
                    return fm.get_metric(obj)
            raise InvalidFieldException("Invalid metric name: %s" % obj)

        if isinstance(obj, dict):
            metric = AdHocMetric.create(obj)
            assert not self.has_metric(metric.name, adhoc_fms=adhoc_fms), (
                "AdHocMetric can not use name of an existing metric: %s" % metric.name
            )
            metric.check_formula_fields(self, adhoc_fms=adhoc_fms)
            return metric

        raise InvalidFieldException("Invalid metric object: %s" % obj)

    def get_dimension(self, obj, adhoc_fms=None):
        if isinstance(obj, str):
            if self._directly_has_dimension(obj):
                return getattr(self, self.dimensions_attr)[obj]
            for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
                if fm.has_dimension(obj):
                    return fm.get_dimension(obj)
            raise InvalidFieldException("Invalid dimension name: %s" % obj)

        if isinstance(obj, dict):
            raise InvalidFieldException("AdHocDimensions are not currently supported")
            # dim = AdHocDimension.create(obj)
            # assert not self.has_dimension(dim.name, adhoc_fms=adhoc_fms), (
            #     "AdHocDimension can not use name of an existing dimension: %s"
            #     % dim.name
            # )
            # return dim

        raise InvalidFieldException("Invalid dimension object: %s" % obj)

    def get_field(self, obj, adhoc_fms=None):
        if isinstance(obj, str):
            if self.has_metric(obj, adhoc_fms=adhoc_fms):
                return self.get_metric(obj, adhoc_fms=adhoc_fms)
            if self.has_dimension(obj, adhoc_fms=adhoc_fms):
                return self.get_dimension(obj, adhoc_fms=adhoc_fms)
            raise InvalidFieldException("Invalid field name: %s" % obj)

        # TODO: should this be allowed?
        if isinstance(obj, dict):
            field = AdHocField.create(obj)
            assert not self.has_field(field.name, adhoc_fms=adhoc_fms), (
                "AdHocField can not use name of an existing field: %s" % field.name
            )
            return field

        raise InvalidFieldException("Invalid field object: %s" % obj)

    def get_metrics(self, adhoc_fms=None):
        metrics = {}
        metrics.update(getattr(self, self.metrics_attr))
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            fm_metrics = fm.get_metrics()
            metrics.update(fm_metrics)
        return metrics

    def get_dimensions(self, adhoc_fms=None):
        dimensions = {}
        dimensions.update(getattr(self, self.dimensions_attr))
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            fm_dimensions = fm.get_dimensions()
            dimensions.update(fm_dimensions)
        return dimensions

    def get_fields(self, adhoc_fms=None):
        fields = {}
        fields.update(getattr(self, self.metrics_attr))
        fields.update(getattr(self, self.dimensions_attr))
        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            fm_fields = fm.get_fields()
            fields.update(fm_fields)
        return fields

    def get_metric_names(self, adhoc_fms=None):
        return set(self.get_metrics(adhoc_fms=adhoc_fms).keys())

    def get_dimension_names(self, adhoc_fms=None):
        return set(self.get_dimensions(adhoc_fms=adhoc_fms).keys())

    def get_field_names(self, adhoc_fms=None):
        return set(self.get_fields(adhoc_fms=adhoc_fms).keys())

    def add_metric(self, metric, force=False):
        if self.has_dimension(metric.name):
            raise InvalidFieldException(
                "Trying to add metric with same name as a dimension: %s" % metric.name
            )
        if (not force) and self.has_metric(metric.name):
            warn("Metric %s already exists on %s" % (metric.name, self))
            return
        getattr(self, self.metrics_attr)[metric.name] = metric

    def add_dimension(self, dimension, force=False):
        if self.has_metric(dimension.name):
            raise InvalidFieldException(
                "Trying to add dimension with same name as a metric: %s"
                % dimension.name
            )
        if (not force) and self.has_dimension(dimension.name):
            warn("Dimension %s already exists on %s" % (dimension.name, self))
            return
        getattr(self, self.dimensions_attr)[dimension.name] = dimension

    def populate_global_fields(self, config, force=False):
        formula_metrics = []
        formula_dims = []

        for metric_def in config.get("metrics", []):
            if isinstance(metric_def, dict):
                schema = MetricConfigSchema()
                metric_def = schema.load(metric_def)
                metric = create_metric(metric_def)
            else:
                assert isinstance(
                    metric_def, Metric
                ), "Metric definition must be a dict-like object or a Metric object"
                metric = metric_def

            if isinstance(metric, FormulaMetric):
                formula_metrics.append(metric)  # These get added later
            else:
                self.add_metric(metric, force=force)

        for dim_def in config.get("dimensions", []):
            if isinstance(dim_def, dict):
                schema = DimensionConfigSchema()
                dim_def = schema.load(dim_def)
                dim = create_dimension(dim_def)
            else:
                assert isinstance(
                    dim_def, Dimension
                ), "Dimension definition must be a dict-like object or a Dimension object"
                dim = dim_def

            if isinstance(dim, FormulaDimension):
                # formula_dims.append(dim) # These get added later
                raise InvalidFieldException(
                    "FormulaDimensions are not currently supported"
                )
            else:
                self.add_dimension(dim, force=force)

        # Defer formulas so params can be checked against existing fields
        for metric in formula_metrics:
            metric.check_formula_fields(self)
            self.add_metric(metric, force=force)

        for dim in formula_dims:
            dim.check_formula_fields(self)
            self.add_dimension(dim, force=force)

    def find_field_sources(self, field, adhoc_fms=None):
        sources = []
        if self._directly_has_field(field):
            sources.append(self)

        for fm in self.get_field_managers(adhoc_fms=adhoc_fms):
            if fm._directly_has_field(field):
                sources.append(fm)
        return sources


def get_table_metrics(fm, table, adhoc_fms=None):
    metrics = set()
    for col in table.c:
        if not (getattr(col, "zillion", None) and col.zillion.active):
            continue
        for field in col.zillion.get_field_names():
            if fm.has_metric(field, adhoc_fms=adhoc_fms):
                metrics.add(field)
    return metrics


def get_table_dimensions(fm, table, adhoc_fms=None):
    dims = set()
    for col in table.c:
        if not (getattr(col, "zillion", None) and col.zillion.active):
            continue
        for field in col.zillion.get_field_names():
            if fm.has_dimension(field, adhoc_fms=adhoc_fms):
                dims.add(field)
    return dims


def get_table_fields(table):
    fields = set()
    for col in table.c:
        if not (getattr(col, "zillion", None) and col.zillion.active):
            continue
        for field in col.zillion.get_field_names():
            fields.add(field)
    return fields


def get_table_field_column(table, field_name):
    for col in table.c:
        if not (getattr(col, "zillion", None) and col.zillion.active):
            continue
        for field in col.zillion.get_field_names():
            if field == field_name:
                return col
    assert False, "Field %s inactive or not found in table %s" % (
        field_name,
        table.fullname,
    )



"""
Useful reference around SQLAlchemy types by dialect:
https://github.com/zzzeek/sqlalchemy/blob/master/lib/sqlalchemy/dialects/type_migration_guidelines.txt
"""

import ast
import re

import sqlalchemy as sa
from sqlalchemy.dialects.mysql import dialect as mysql_dialect
from sqlalchemy.dialects.postgresql import dialect as postgresql_dialect
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.engine import reflection
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression as exp
import sqlparse as sp
from tlbx import dbg, st, get_class_vars, get_string_format_args

from zillion.core import AggregationTypes

DIGIT_THRESHOLD_FOR_AVG_AGGR = 1

INTEGER_SA_TYPES = [
    sa.BigInteger,
    sa.BIGINT,
    sa.Integer,
    sa.INT,
    sa.INTEGER,
    sa.SmallInteger,
    sa.SMALLINT,
]

FLOAT_SA_TYPES = [
    sa.DECIMAL,
    sa.Float,
    sa.FLOAT,
    sa.Numeric,
    sa.NUMERIC,
    sa.REAL,
    sa.dialects.postgresql.DOUBLE_PRECISION,
    sa.dialects.postgresql.MONEY,
]

NUMERIC_SA_TYPES = INTEGER_SA_TYPES + FLOAT_SA_TYPES

DATETIME_SA_TYPES = [sa.DateTime, sa.DATETIME, sa.Time, sa.TIME, sa.TIMESTAMP]

DATE_SA_TYPES = [sa.Date, sa.DATE]

DATE_HIERARCHY = [
    "year",
    "quarter",
    "quarter_of_year",
    "month",
    "month_name",
    "month_of_year",
    "date",
    "day_name",
    "day_of_week",
    "day_of_month",
    "day_of_year",
    "hour",
    "hour_of_day",
    "minute",
    "minute_of_hour",
    "datetime",
    "unixtime",
]

# Somewhat adhering to ISO 8601, but ignoring the "T" between the date/time
# and not including timezone offsets for now because zillion assumes
# everything is in the same timezone (or the datasource formulas take care of
# aligning timezones).
DIALECT_DATE_CONVERSIONS = {
    "sqlite": {
        "year": "cast(strftime('%Y', {}) as integer)",
        "quarter": "strftime('%Y', {}) || '-Q' || ((cast(strftime('%m', {}) as integer) + 2) / 3)",  # 2020-Q1
        "quarter_of_year": "(cast(strftime('%m', {}) as integer) + 2) / 3",
        "month": "strftime('%Y-%m', {})",
        "month_name": (
            "CASE strftime('%m', {}) "
            "WHEN '01' THEN 'January' "
            "WHEN '02' THEN 'February' "
            "WHEN '03' THEN 'March' "
            "WHEN '04' THEN 'April' "
            "WHEN '05' THEN 'May' "
            "WHEN '06' THEN 'June' "
            "WHEN '07' THEN 'July' "
            "WHEN '08' THEN 'August' "
            "WHEN '09' THEN 'September' "
            "WHEN '10' THEN 'October' "
            "WHEN '11' THEN 'November' "
            "WHEN '12' THEN 'December' "
            "ELSE NULL "
            "END"
        ),
        "month_of_year": "cast(strftime('%m', {}) as integer)",
        "date": "strftime('%Y-%m-%d', {})",
        "day_name": (
            "CASE cast(strftime('%w', {}) as integer) "
            "WHEN 0 THEN 'Sunday' "
            "WHEN 1 THEN 'Monday' "
            "WHEN 2 THEN 'Tuesday' "
            "WHEN 3 THEN 'Wednesday' "
            "WHEN 4 THEN 'Thursday' "
            "WHEN 5 THEN 'Friday' "
            "WHEN 6 THEN 'Saturday' "
            "ELSE NULL "
            "END"
        ),
        "day_of_week": "(cast(strftime('%w', {}) as integer) + 6) % 7 + 1",  # Convert to Monday = 1
        "day_of_month": "cast(strftime('%d', {}) as integer)",
        "day_of_year": "cast(strftime('%j', {}) as integer)",
        "hour": "strftime('%Y-%m-%d %H:00:00', {})",
        "hour_of_day": "cast(strftime('%H', {}) as integer)",
        "minute": "strftime('%Y-%m-%d %H:%M:00', {})",
        "minute_of_hour": "cast(strftime('%M', {}) as integer)",
        "datetime": "strftime('%Y-%m-%d %H:%M:%S', {})",
        "unixtime": "cast(strftime('%s', {}) as integer)",
    },
    "mysql": {
        "year": "EXTRACT(YEAR FROM {})",
        "quarter": "CONCAT(YEAR({}), '-Q', QUARTER({}))",
        "quarter_of_year": "EXTRACT(QUARTER FROM {})",
        "month": "DATE_FORMAT({}, '%Y-%m')",
        "month_name": "MONTHNAME({})",
        "month_of_year": "EXTRACT(MONTH FROM {})",
        "date": "DATE_FORMAT({}, '%Y-%m-%d')",
        "day_name": "DAYNAME({})",
        "day_of_week": "WEEKDAY({}) + 1",  # Monday = 1
        "day_of_month": "EXTRACT(DAY FROM {})",
        "day_of_year": "DAYOFYEAR({})",
        "hour": "DATE_FORMAT({}, '%Y-%m-%d %H:00:00')",
        "hour_of_day": "EXTRACT(HOUR FROM {})",
        "minute": "DATE_FORMAT({}, '%Y-%m-%d %H:%i:00')",
        "minute_of_hour": "EXTRACT(MINUTE FROM {})",
        "datetime": "DATE_FORMAT({}, '%Y-%m-%d %H:%i:%S')",
        "unixtime": "UNIX_TIMESTAMP({})",
    },
    "postgresql": {
        "year": "EXTRACT(YEAR FROM {})",
        "quarter": "TO_CHAR({}, 'FMYYYY-\"Q\"Q')",
        "quarter_of_year": "EXTRACT(QUARTER FROM {})",
        "month": "TO_CHAR({}, 'FMYYYY-MM')",
        "month_name": "TO_CHAR({}, 'FMMonth')",
        "month_of_year": "EXTRACT(MONTH FROM {})",
        "date": "TO_CHAR({}, 'FMYYYY-MM-DD')",
        "day_name": "TO_CHAR({}, 'FMDay')",
        "day_of_week": "EXTRACT(ISODOW FROM {})",  # Monday = 1
        "day_of_month": "EXTRACT(DAY FROM {})",
        "day_of_year": "EXTRACT(DOY FROM {})",
        "hour": "TO_CHAR({}, 'FMYYYY-MM-DD HH24:00:00')",
        "hour_of_day": "EXTRACT(HOUR FROM {})",
        "minute": "TO_CHAR({}, 'FMYYYY-MM-DD HH24:MI:00')",
        "minute_of_hour": "EXTRACT(MINUTE FROM {})",
        "datetime": "TO_CHAR({}, 'FMYYYY-MM-DD HH24:MI:SS')",
        "unixtime": "EXTRACT(epoch from {})",
    },
}

TYPE_ALLOWED_CONVERSIONS = {
    sa.DateTime: {
        "allowed_conversions": DATE_HIERARCHY,
        "dialect_conversions": DIALECT_DATE_CONVERSIONS,
    },
    sa.DATETIME: {
        "allowed_conversions": DATE_HIERARCHY,
        "dialect_conversions": DIALECT_DATE_CONVERSIONS,
    },
    sa.TIMESTAMP: {
        "allowed_conversions": DATE_HIERARCHY,
        "dialect_conversions": DIALECT_DATE_CONVERSIONS,
    },
    sa.Date: {
        "allowed_conversions": DATE_HIERARCHY[0 : DATE_HIERARCHY.index("hour")],
        "dialect_conversions": DIALECT_DATE_CONVERSIONS,
    },
    sa.DATE: {
        "allowed_conversions": DATE_HIERARCHY[0 : DATE_HIERARCHY.index("hour")],
        "dialect_conversions": DIALECT_DATE_CONVERSIONS,
    },
}

AGGREGATION_SQLA_FUNC_MAP = {
    AggregationTypes.AVG: sa.func.avg,
    AggregationTypes.COUNT: sa.func.count,
    AggregationTypes.COUNT_DISTINCT: lambda x: sa.func.count(sa.distinct(x)),
    AggregationTypes.MIN: sa.func.min,
    AggregationTypes.MAX: sa.func.max,
    AggregationTypes.SUM: sa.func.sum,
}

# This establishes a baseline of schemas to ignore during reflection
DIALECT_IGNORE_SCHEMAS = {
    "mysql": set(["information_schema", "performance_schema", "mysql", "sys"]),
    "postgresql": set(["information_schema", r"pg_(.*)"]),
    # TODO: more support here
}


class InvalidSQLAlchemyTypeString(Exception):
    pass


def get_conversions_for_type(coltype):
    for basetype, convs in TYPE_ALLOWED_CONVERSIONS.items():
        if issubclass(coltype, basetype):
            return convs
    return None


def get_dialect_type_conversions(dialect, column):
    coltype = type(column.type)
    conv_info = get_conversions_for_type(coltype)
    if not conv_info:
        return []

    results = []
    allowed = conv_info["allowed_conversions"]
    convs = conv_info["dialect_conversions"]

    for field in allowed:
        conv = convs[dialect].get(field, None)
        if not conv:
            continue
        format_args = get_string_format_args(conv)
        assert not any([x != "" for x in format_args]), (
            "Field conversion has non-named format arguments: %s" % conv
        )
        if format_args:
            conv = conv.format(*[column_fullname(column) for i in format_args])
        results.append((field, conv))

    return results


def contains_sql_keywords(sql):
    if isinstance(sql, str):
        sql = sp.parse(sql)

    for token in sql:
        if token.ttype in (sp.tokens.DML, sp.tokens.DDL, sp.tokens.CTE):
            return True

        if isinstance(token, sp.sql.TokenList):
            token_result = contains_sql_keywords(token)
            if token_result:
                return True
    return False


def contains_aggregation(sql):
    if isinstance(sql, str):
        sql = sp.parse(sql)

    aggr_types = {x.lower() for x in get_class_vars(AggregationTypes)}

    for token in sql:
        if isinstance(token, sp.sql.Function):
            name = token.get_name()
            # NOTE: If AggregationTypes naming is changed, this could fail
            if name.lower() in aggr_types:
                return True
        if isinstance(token, sp.sql.TokenList):
            token_result = contains_aggregation(token)
            if token_result:
                return True
    return False


def type_string_to_sa_type(type_string):
    # This only checks the top level sqlalchemy module for matching type
    # classes. Therefore you can not specify dialect-specific types at this
    # time.
    parts = type_string.split("(")
    type_args = []
    if len(parts) > 1:
        assert len(parts) == 2, "Unable to parse type string: %s" % type_string
        type_args = ast.literal_eval(parts[1].rstrip(")") + ",")
    type_name = parts[0]
    type_cls = getattr(sa.types, type_name, None)
    if not type_cls:
        raise InvalidSQLAlchemyTypeString(
            "Could not find matching type for %s" % type_name
        )
    return type_cls(*type_args)


def infer_aggregation_and_rounding(column):
    if isinstance(column.type, tuple(INTEGER_SA_TYPES)):
        return AggregationTypes.SUM, 0
    if isinstance(column.type, tuple(FLOAT_SA_TYPES)):
        rounding = column.type.scale
        precision = column.type.precision
        if rounding is None and precision is None:
            aggregation = AggregationTypes.SUM
        else:
            whole_digits = precision - rounding
            if whole_digits <= DIGIT_THRESHOLD_FOR_AVG_AGGR:
                aggregation = AggregationTypes.AVG
            else:
                aggregation = AggregationTypes.SUM
        return aggregation, rounding
    assert False, "Column %s is not a numeric type" % column


def aggregation_to_sqla_func(aggregation):
    return AGGREGATION_SQLA_FUNC_MAP[aggregation]


def is_probably_metric(column, formula=None):
    if formula and contains_aggregation(formula):
        return True
    if not isinstance(column.type, tuple(NUMERIC_SA_TYPES)):
        return False
    if column.primary_key:
        return False
    if column.name.endswith("_id") or column.name.endswith("Id") or column.name == "id":
        return False
    return True


def sqla_compile(expr):
    return str(expr.compile(compile_kwargs={"literal_binds": True}))


def printexpr(expr):
    print(sqla_compile(expr))


def column_fullname(column, prefix=None):
    name = "%s.%s" % (column.table.fullname, column.name)
    if prefix:
        name = prefix + "." + name
    return name


def get_sqla_clause(column, criterion, negate=False):
    """
    TODO: postgresql like is case sensitive, but mysql like is not
    - postgres also supports ilike to specify case insensitive
    OPTION: look at dialect to determine function
    """
    field, op, values = criterion
    op = op.lower()
    if not isinstance(values, (list, tuple)):
        values = [values]

    use_or = True
    has_null = False
    for v in values:
        # len() check is to avoid calling string.lower() on huge strings
        if v is None or (isinstance(v, str) and len(v) == 4 and v.lower() == "null"):
            has_null = True

    if op == "=":
        clauses = [column == v if v is not None else column._is(None) for v in values]
    elif op == "!=":
        clauses = [column != v if v is not None else column.isnot(None) for v in values]
    elif op == ">":
        clauses = [column > v for v in values]
    elif op == "<":
        clauses = [column < v for v in values]
    elif op == ">=":
        clauses = [column >= v for v in values]
    elif op == "<=":
        clauses = [column <= v for v in values]
    elif op == "in":
        if has_null:
            clauses = [
                column == v if v is not None else column._is(None) for v in values
            ]
        else:
            clauses = [column.in_(values)]
    elif op == "not in":
        use_or = False
        if has_null:
            clauses = [
                column != v if v is not None else column.isnot(None) for v in values
            ]
        else:
            clauses = [sa.not_(column.in_(values))]
    elif op == "between":
        clauses = []
        for value in values:
            assert len(value) == 2, "Between clause value must have length of 2"
            clauses.append(column.between(value[0], value[1]))
    elif op == "not between":
        clauses = []
        for value in values:
            assert len(value) == 2, "Between clause value must have length of 2"
            clauses.append(sa.not_(column.between(value[0], value[1])))
    elif op == "like":
        clauses = [column.like(v) for v in values]
    elif op == "not like":
        use_or = False
        clauses = [sa.not_(column.like(v)) for v in values]
    else:
        assert False, "Invalid criterion operand: %s" % op

    if use_or:
        clause = sa.or_(*clauses)
    else:
        clause = sa.and_(*clauses)

    if negate:
        clause = sa.not_(clause)

    return clause


# https://github.com/sqlalchemy/sqlalchemy/wiki/CompiledComments
def comment(self, comment):
    self._added_comment = comment
    return self


exp.ClauseElement.comment = comment
exp.ClauseElement._added_comment = None


def _compile_element(elem, prepend_newline=False):
    @compiles(elem)
    def add_comment(element, compiler, **kw):
        meth = getattr(compiler, "visit_%s" % element.__visit_name__)
        text = meth(element, **kw)
        if element._added_comment:
            # Modified this line to not add newline
            text = "-- %s\n" % element._added_comment + text
        elif prepend_newline:
            text = "\n" + text
        return text


_compile_element(exp.Case)
_compile_element(exp.Label, True)
_compile_element(exp.ColumnClause)
_compile_element(exp.Join)
_compile_element(exp.Select)
_compile_element(exp.Alias)
_compile_element(exp.Exists)


def get_schemas(engine):
    insp = reflection.Inspector.from_engine(engine)
    return insp.get_schema_names()


# -------- Some DB-specific stuff


def to_mysql_type(type):
    return type.compile(dialect=mysql_dialect())


def to_postgresql_type(type):
    return type.compile(dialect=postgresql_dialect())


def to_sqlite_type(type):
    return type.compile(dialect=sqlite_dialect())


def filter_dialect_schemas(schemas, dialect):
    ignores = DIALECT_IGNORE_SCHEMAS.get(dialect, None)
    if not ignores:
        return schemas

    final = []
    for schema in schemas:
        add = True
        for ignore in ignores:
            if re.match(ignore, schema):
                add = False
                break
        if add:
            final.append(schema)

    return final


def get_postgres_schemas(conn):
    qr = conn.execute(
        sa.text(
            "SELECT schema_name FROM information_schema.schemata "
            "WHERE schema_name not LIKE 'pg_%' and schema_name != 'information_schema'"
        )
    )
    return [x["schema_name"] for x in qr.fetchall()]


def get_postgres_pid(conn):
    qr = conn.execute("select pg_backend_pid()")
    pid = qr.fetchone()[0]
    return pid
